"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any, Final, Hashable, Iterator, Literal, TYPE_CHECKING
from pandas._typing import AnyArrayLike, ArrayLike, DtypeArg, FilePath, Shape, npt
from pandas.util._decorators import cache_readonly
from pandas import DataFrame, DatetimeIndex, Index, MultiIndex, Series
from pandas.core.api import Int64Index
from pandas.core.computation.pytables import PyTablesExpr
from tables import Col, File, Node

"""
High level interface to PyTables for reading and writing pandas data structures
to disk
"""
if TYPE_CHECKING:
    ...
_version = ...
_default_encoding = ...
Term = PyTablesExpr
incompatibility_doc: Final = ...
attribute_conflict_doc: Final = ...
performance_doc: Final = ...
_FORMAT_MAP = ...
_AXES_MAP = ...
dropna_doc: Final = ...
format_doc: Final = ...
_table_mod = ...
_table_file_open_policy_is_strict = ...
def to_hdf(path_or_buf: FilePath | HDFStore, key: str, value: DataFrame | Series, mode: str = ..., complevel: int | None = ..., complib: str | None = ..., append: bool = ..., format: str | None = ..., index: bool = ..., min_itemsize: int | dict[str, int] | None = ..., nan_rep=..., dropna: bool | None = ..., data_columns: Literal[True] | list[str] | None = ..., errors: str = ..., encoding: str = ...) -> None:
    """store this object, close it if we opened it"""
    ...

def read_hdf(path_or_buf: FilePath | HDFStore, key=..., mode: str = ..., errors: str = ..., where: str | list | None = ..., start: int | None = ..., stop: int | None = ..., columns: list[str] | None = ..., iterator: bool = ..., chunksize: int | None = ..., **kwargs): # -> TableIterator:
    """
    Read from the store, close it if we opened it.

    Retrieve pandas object stored in file, optionally based on where
    criteria.

    .. warning::

       Pandas uses PyTables for reading and writing HDF5 files, which allows
       serializing object-dtype data with pickle when using the "fixed" format.
       Loading pickled data received from untrusted sources can be unsafe.

       See: https://docs.python.org/3/library/pickle.html for more.

    Parameters
    ----------
    path_or_buf : str, path object, pandas.HDFStore
        Any valid string path is acceptable. Only supports the local file system,
        remote URLs and file-like objects are not supported.

        If you want to pass in a path object, pandas accepts any
        ``os.PathLike``.

        Alternatively, pandas accepts an open :class:`pandas.HDFStore` object.

    key : object, optional
        The group identifier in the store. Can be omitted if the HDF file
        contains a single pandas object.
    mode : {'r', 'r+', 'a'}, default 'r'
        Mode to use when opening the file. Ignored if path_or_buf is a
        :class:`pandas.HDFStore`. Default is 'r'.
    errors : str, default 'strict'
        Specifies how encoding and decoding errors are to be handled.
        See the errors argument for :func:`open` for a full list
        of options.
    where : list, optional
        A list of Term (or convertible) objects.
    start : int, optional
        Row number to start selection.
    stop  : int, optional
        Row number to stop selection.
    columns : list, optional
        A list of columns names to return.
    iterator : bool, optional
        Return an iterator object.
    chunksize : int, optional
        Number of rows to include in an iteration when using an iterator.
    **kwargs
        Additional keyword arguments passed to HDFStore.

    Returns
    -------
    item : object
        The selected object. Return type depends on the object stored.

    See Also
    --------
    DataFrame.to_hdf : Write a HDF file from a DataFrame.
    HDFStore : Low-level access to HDF files.

    Examples
    --------
    >>> df = pd.DataFrame([[1, 1.0, 'a']], columns=['x', 'y', 'z'])  # doctest: +SKIP
    >>> df.to_hdf('./store.h5', 'data')  # doctest: +SKIP
    >>> reread = pd.read_hdf('./store.h5')  # doctest: +SKIP
    """
    ...

class HDFStore:
    """
    Dict-like IO interface for storing pandas objects in PyTables.

    Either Fixed or Table format.

    .. warning::

       Pandas uses PyTables for reading and writing HDF5 files, which allows
       serializing object-dtype data with pickle when using the "fixed" format.
       Loading pickled data received from untrusted sources can be unsafe.

       See: https://docs.python.org/3/library/pickle.html for more.

    Parameters
    ----------
    path : str
        File path to HDF5 file.
    mode : {'a', 'w', 'r', 'r+'}, default 'a'

        ``'r'``
            Read-only; no data can be modified.
        ``'w'``
            Write; a new file is created (an existing file with the same
            name would be deleted).
        ``'a'``
            Append; an existing file is opened for reading and writing,
            and if the file does not exist it is created.
        ``'r+'``
            It is similar to ``'a'``, but the file must already exist.
    complevel : int, 0-9, default None
        Specifies a compression level for data.
        A value of 0 or None disables compression.
    complib : {'zlib', 'lzo', 'bzip2', 'blosc'}, default 'zlib'
        Specifies the compression library to be used.
        As of v0.20.2 these additional compressors for Blosc are supported
        (default if no compressor specified: 'blosc:blosclz'):
        {'blosc:blosclz', 'blosc:lz4', 'blosc:lz4hc', 'blosc:snappy',
         'blosc:zlib', 'blosc:zstd'}.
        Specifying a compression library which is not available issues
        a ValueError.
    fletcher32 : bool, default False
        If applying compression use the fletcher32 checksum.
    **kwargs
        These parameters will be passed to the PyTables open_file method.

    Examples
    --------
    >>> bar = pd.DataFrame(np.random.randn(10, 4))
    >>> store = pd.HDFStore('test.h5')
    >>> store['foo'] = bar   # write to HDF5
    >>> bar = store['foo']   # retrieve
    >>> store.close()

    **Create or load HDF5 file in-memory**

    When passing the `driver` option to the PyTables open_file method through
    **kwargs, the HDF5 file is loaded or created in-memory and will only be
    written when closed:

    >>> bar = pd.DataFrame(np.random.randn(10, 4))
    >>> store = pd.HDFStore('test.h5', driver='H5FD_CORE')
    >>> store['foo'] = bar
    >>> store.close()   # only now, data is written to disk
    """
    _handle: File | None
    _mode: str
    _complevel: int
    _fletcher32: bool
    def __init__(self, path, mode: str = ..., complevel: int | None = ..., complib=..., fletcher32: bool = ..., **kwargs) -> None:
        ...
    
    def __fspath__(self) -> str:
        ...
    
    @property
    def root(self):
        """return the root node"""
        ...
    
    @property
    def filename(self) -> str:
        ...
    
    def __getitem__(self, key: str):
        ...
    
    def __setitem__(self, key: str, value) -> None:
        ...
    
    def __delitem__(self, key: str) -> None:
        ...
    
    def __getattr__(self, name: str):
        """allow attribute access to get stores"""
        ...
    
    def __contains__(self, key: str) -> bool:
        """
        check for existence of this key
        can match the exact pathname or the pathnm w/o the leading '/'
        """
        ...
    
    def __len__(self) -> int:
        ...
    
    def __repr__(self) -> str:
        ...
    
    def __enter__(self) -> HDFStore:
        ...
    
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...
    
    def keys(self, include: str = ...) -> list[str]:
        """
        Return a list of keys corresponding to objects stored in HDFStore.

        Parameters
        ----------

        include : str, default 'pandas'
                When kind equals 'pandas' return pandas objects.
                When kind equals 'native' return native HDF5 Table objects.

                .. versionadded:: 1.1.0

        Returns
        -------
        list
            List of ABSOLUTE path-names (e.g. have the leading '/').

        Raises
        ------
        raises ValueError if kind has an illegal value
        """
        ...
    
    def __iter__(self) -> Iterator[str]:
        ...
    
    def items(self) -> Iterator[tuple[str, list]]:
        """
        iterate on key->group
        """
        ...
    
    def iteritems(self): # -> Generator[tuple[str, list[Unknown]], None, None]:
        """
        iterate on key->group
        """
        ...
    
    def open(self, mode: str = ..., **kwargs) -> None:
        """
        Open the file in the specified mode

        Parameters
        ----------
        mode : {'a', 'w', 'r', 'r+'}, default 'a'
            See HDFStore docstring or tables.open_file for info about modes
        **kwargs
            These parameters will be passed to the PyTables open_file method.
        """
        ...
    
    def close(self) -> None:
        """
        Close the PyTables file handle
        """
        ...
    
    @property
    def is_open(self) -> bool:
        """
        return a boolean indicating whether the file is open
        """
        ...
    
    def flush(self, fsync: bool = ...) -> None:
        """
        Force all buffered modifications to be written to disk.

        Parameters
        ----------
        fsync : bool (default False)
          call ``os.fsync()`` on the file handle to force writing to disk.

        Notes
        -----
        Without ``fsync=True``, flushing may not guarantee that the OS writes
        to disk. With fsync, the operation will block until the OS claims the
        file has been written; however, other caching layers may still
        interfere.
        """
        ...
    
    def get(self, key: str):
        """
        Retrieve pandas object stored in file.

        Parameters
        ----------
        key : str

        Returns
        -------
        object
            Same type as object stored in file.
        """
        ...
    
    def select(self, key: str, where=..., start=..., stop=..., columns=..., iterator=..., chunksize=..., auto_close: bool = ...): # -> TableIterator:
        """
        Retrieve pandas object stored in file, optionally based on where criteria.

        .. warning::

           Pandas uses PyTables for reading and writing HDF5 files, which allows
           serializing object-dtype data with pickle when using the "fixed" format.
           Loading pickled data received from untrusted sources can be unsafe.

           See: https://docs.python.org/3/library/pickle.html for more.

        Parameters
        ----------
        key : str
            Object being retrieved from file.
        where : list or None
            List of Term (or convertible) objects, optional.
        start : int or None
            Row number to start selection.
        stop : int, default None
            Row number to stop selection.
        columns : list or None
            A list of columns that if not None, will limit the return columns.
        iterator : bool or False
            Returns an iterator.
        chunksize : int or None
            Number or rows to include in iteration, return an iterator.
        auto_close : bool or False
            Should automatically close the store when finished.

        Returns
        -------
        object
            Retrieved object from file.
        """
        ...
    
    def select_as_coordinates(self, key: str, where=..., start: int | None = ..., stop: int | None = ...): # -> Index | Literal[False]:
        """
        return the selection as an Index

        .. warning::

           Pandas uses PyTables for reading and writing HDF5 files, which allows
           serializing object-dtype data with pickle when using the "fixed" format.
           Loading pickled data received from untrusted sources can be unsafe.

           See: https://docs.python.org/3/library/pickle.html for more.


        Parameters
        ----------
        key : str
        where : list of Term (or convertible) objects, optional
        start : integer (defaults to None), row number to start selection
        stop  : integer (defaults to None), row number to stop selection
        """
        ...
    
    def select_column(self, key: str, column: str, start: int | None = ..., stop: int | None = ...): # -> Series | Literal[False]:
        """
        return a single column from the table. This is generally only useful to
        select an indexable

        .. warning::

           Pandas uses PyTables for reading and writing HDF5 files, which allows
           serializing object-dtype data with pickle when using the "fixed" format.
           Loading pickled data received from untrusted sources can be unsafe.

           See: https://docs.python.org/3/library/pickle.html for more.

        Parameters
        ----------
        key : str
        column : str
            The column of interest.
        start : int or None, default None
        stop : int or None, default None

        Raises
        ------
        raises KeyError if the column is not found (or key is not a valid
            store)
        raises ValueError if the column can not be extracted individually (it
            is part of a data block)

        """
        ...
    
    def select_as_multiple(self, keys, where=..., selector=..., columns=..., start=..., stop=..., iterator=..., chunksize=..., auto_close: bool = ...): # -> TableIterator:
        """
        Retrieve pandas objects from multiple tables.

        .. warning::

           Pandas uses PyTables for reading and writing HDF5 files, which allows
           serializing object-dtype data with pickle when using the "fixed" format.
           Loading pickled data received from untrusted sources can be unsafe.

           See: https://docs.python.org/3/library/pickle.html for more.

        Parameters
        ----------
        keys : a list of the tables
        selector : the table to apply the where criteria (defaults to keys[0]
            if not supplied)
        columns : the columns I want back
        start : integer (defaults to None), row number to start selection
        stop  : integer (defaults to None), row number to stop selection
        iterator : bool, return an iterator, default False
        chunksize : nrows to include in iteration, return an iterator
        auto_close : bool, default False
            Should automatically close the store when finished.

        Raises
        ------
        raises KeyError if keys or selector is not found or keys is empty
        raises TypeError if keys is not a list or tuple
        raises ValueError if the tables are not ALL THE SAME DIMENSIONS
        """
        ...
    
    def put(self, key: str, value: DataFrame | Series, format=..., index=..., append=..., complib=..., complevel: int | None = ..., min_itemsize: int | dict[str, int] | None = ..., nan_rep=..., data_columns: Literal[True] | list[str] | None = ..., encoding=..., errors: str = ..., track_times: bool = ..., dropna: bool = ...) -> None:
        """
        Store object in HDFStore.

        Parameters
        ----------
        key : str
        value : {Series, DataFrame}
        format : 'fixed(f)|table(t)', default is 'fixed'
            Format to use when storing object in HDFStore. Value can be one of:

            ``'fixed'``
                Fixed format.  Fast writing/reading. Not-appendable, nor searchable.
            ``'table'``
                Table format.  Write as a PyTables Table structure which may perform
                worse but allow more flexible operations like searching / selecting
                subsets of the data.
        index : bool, default True
            Write DataFrame index as a column.
        append : bool, default False
            This will force Table format, append the input data to the existing.
        data_columns : list of columns or True, default None
            List of columns to create as data columns, or True to use all columns.
            See `here
            <https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#query-via-data-columns>`__.
        encoding : str, default None
            Provide an encoding for strings.
        track_times : bool, default True
            Parameter is propagated to 'create_table' method of 'PyTables'.
            If set to False it enables to have the same h5 files (same hashes)
            independent on creation time.
        dropna : bool, default False, optional
            Remove missing values.

            .. versionadded:: 1.1.0
        """
        ...
    
    def remove(self, key: str, where=..., start=..., stop=...) -> None:
        """
        Remove pandas object partially by specifying the where condition

        Parameters
        ----------
        key : str
            Node to remove or delete rows from
        where : list of Term (or convertible) objects, optional
        start : integer (defaults to None), row number to start selection
        stop  : integer (defaults to None), row number to stop selection

        Returns
        -------
        number of rows removed (or None if not a Table)

        Raises
        ------
        raises KeyError if key is not a valid store

        """
        ...
    
    def append(self, key: str, value: DataFrame | Series, format=..., axes=..., index=..., append=..., complib=..., complevel: int | None = ..., columns=..., min_itemsize: int | dict[str, int] | None = ..., nan_rep=..., chunksize=..., expectedrows=..., dropna: bool | None = ..., data_columns: Literal[True] | list[str] | None = ..., encoding=..., errors: str = ...) -> None:
        """
        Append to Table in file.

        Node must already exist and be Table format.

        Parameters
        ----------
        key : str
        value : {Series, DataFrame}
        format : 'table' is the default
            Format to use when storing object in HDFStore.  Value can be one of:

            ``'table'``
                Table format. Write as a PyTables Table structure which may perform
                worse but allow more flexible operations like searching / selecting
                subsets of the data.
        index : bool, default True
            Write DataFrame index as a column.
        append       : bool, default True
            Append the input data to the existing.
        data_columns : list of columns, or True, default None
            List of columns to create as indexed data columns for on-disk
            queries, or True to use all columns. By default only the axes
            of the object are indexed. See `here
            <https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#query-via-data-columns>`__.
        min_itemsize : dict of columns that specify minimum str sizes
        nan_rep      : str to use as str nan representation
        chunksize    : size to chunk the writing
        expectedrows : expected TOTAL row size of this table
        encoding     : default None, provide an encoding for str
        dropna : bool, default False, optional
            Do not write an ALL nan row to the store settable
            by the option 'io.hdf.dropna_table'.

        Notes
        -----
        Does *not* check if data being appended overlaps with existing
        data in the table, so be careful
        """
        ...
    
    def append_to_multiple(self, d: dict, value, selector, data_columns=..., axes=..., dropna=..., **kwargs) -> None:
        """
        Append to multiple tables

        Parameters
        ----------
        d : a dict of table_name to table_columns, None is acceptable as the
            values of one node (this will get all the remaining columns)
        value : a pandas object
        selector : a string that designates the indexable table; all of its
            columns will be designed as data_columns, unless data_columns is
            passed, in which case these are used
        data_columns : list of columns to create as data columns, or True to
            use all columns
        dropna : if evaluates to True, drop rows from all tables if any single
                 row in each table has all NaN. Default False.

        Notes
        -----
        axes parameter is currently not accepted

        """
        ...
    
    def create_table_index(self, key: str, columns=..., optlevel: int | None = ..., kind: str | None = ...) -> None:
        """
        Create a pytables index on the table.

        Parameters
        ----------
        key : str
        columns : None, bool, or listlike[str]
            Indicate which columns to create an index on.

            * False : Do not create any indexes.
            * True : Create indexes on all columns.
            * None : Create indexes on all columns.
            * listlike : Create indexes on the given columns.

        optlevel : int or None, default None
            Optimization level, if None, pytables defaults to 6.
        kind : str or None, default None
            Kind of index, if None, pytables defaults to "medium".

        Raises
        ------
        TypeError: raises if the node is not a table
        """
        ...
    
    def groups(self) -> list:
        """
        Return a list of all the top-level nodes.

        Each node returned is not a pandas storage object.

        Returns
        -------
        list
            List of objects.
        """
        ...
    
    def walk(self, where: str = ...) -> Iterator[tuple[str, list[str], list[str]]]:
        """
        Walk the pytables group hierarchy for pandas objects.

        This generator will yield the group path, subgroups and pandas object
        names for each group.

        Any non-pandas PyTables objects that are not a group will be ignored.

        The `where` group itself is listed first (preorder), then each of its
        child groups (following an alphanumerical order) is also traversed,
        following the same procedure.

        Parameters
        ----------
        where : str, default "/"
            Group where to start walking.

        Yields
        ------
        path : str
            Full path to a group (without trailing '/').
        groups : list
            Names (strings) of the groups contained in `path`.
        leaves : list
            Names (strings) of the pandas objects contained in `path`.
        """
        ...
    
    def get_node(self, key: str) -> Node | None:
        """return the node with the key or None if it does not exist"""
        ...
    
    def get_storer(self, key: str) -> GenericFixed | Table:
        """return the storer object for a key, raise if not in the file"""
        ...
    
    def copy(self, file, mode=..., propindexes: bool = ..., keys=..., complib=..., complevel: int | None = ..., fletcher32: bool = ..., overwrite=...) -> HDFStore:
        """
        Copy the existing store to a new file, updating in place.

        Parameters
        ----------
        propindexes : bool, default True
            Restore indexes in copied file.
        keys : list, optional
            List of keys to include in the copy (defaults to all).
        overwrite : bool, default True
            Whether to overwrite (remove and replace) existing nodes in the new store.
        mode, complib, complevel, fletcher32 same as in HDFStore.__init__

        Returns
        -------
        open file handle of the new store
        """
        ...
    
    def info(self) -> str:
        """
        Print detailed information on the store.

        Returns
        -------
        str
        """
        ...
    


class TableIterator:
    """
    Define the iteration interface on a table

    Parameters
    ----------
    store : HDFStore
    s     : the referred storer
    func  : the function to execute the query
    where : the where of the query
    nrows : the rows to iterate on
    start : the passed start value (default is None)
    stop  : the passed stop value (default is None)
    iterator : bool, default False
        Whether to use the default iterator.
    chunksize : the passed chunking value (default is 100000)
    auto_close : bool, default False
        Whether to automatically close the store at the end of iteration.
    """
    chunksize: int | None
    store: HDFStore
    s: GenericFixed | Table
    def __init__(self, store: HDFStore, s: GenericFixed | Table, func, where, nrows, start=..., stop=..., iterator: bool = ..., chunksize: int | None = ..., auto_close: bool = ...) -> None:
        ...
    
    def __iter__(self): # -> Generator[Unknown, None, None]:
        ...
    
    def close(self) -> None:
        ...
    
    def get_result(self, coordinates: bool = ...): # -> Self@TableIterator:
        ...
    


class IndexCol:
    """
    an index column description class

    Parameters
    ----------
    axis   : axis which I reference
    values : the ndarray like converted values
    kind   : a string description of this type
    typ    : the pytables type
    pos    : the position in the pytables

    """
    is_an_indexable: bool = ...
    is_data_indexable: bool = ...
    _info_fields = ...
    name: str
    cname: str
    def __init__(self, name: str, values=..., kind=..., typ=..., cname: str | None = ..., axis=..., pos=..., freq=..., tz=..., index_name=..., ordered=..., table=..., meta=..., metadata=...) -> None:
        ...
    
    @property
    def itemsize(self) -> int:
        ...
    
    @property
    def kind_attr(self) -> str:
        ...
    
    def set_pos(self, pos: int) -> None:
        """set the position of this column in the Table"""
        ...
    
    def __repr__(self) -> str:
        ...
    
    def __eq__(self, other: Any) -> bool:
        """compare 2 col items"""
        ...
    
    def __ne__(self, other) -> bool:
        ...
    
    @property
    def is_indexed(self) -> bool:
        """return whether I am an indexed column"""
        ...
    
    def convert(self, values: np.ndarray, nan_rep, encoding: str, errors: str) -> tuple[np.ndarray, np.ndarray] | tuple[DatetimeIndex, DatetimeIndex]:
        """
        Convert the data from this selection to the appropriate pandas type.
        """
        ...
    
    def take_data(self): # -> None:
        """return the values"""
        ...
    
    @property
    def attrs(self): # -> Any:
        ...
    
    @property
    def description(self): # -> Any:
        ...
    
    @property
    def col(self): # -> Any | None:
        """return my current col description"""
        ...
    
    @property
    def cvalues(self): # -> None:
        """return my cython values"""
        ...
    
    def __iter__(self):
        ...
    
    def maybe_set_size(self, min_itemsize=...) -> None:
        """
        maybe set a string col itemsize:
            min_itemsize can be an integer or a dict with this columns name
            with an integer size
        """
        ...
    
    def validate_names(self) -> None:
        ...
    
    def validate_and_set(self, handler: AppendableTable, append: bool) -> None:
        ...
    
    def validate_col(self, itemsize=...): # -> Any | None:
        """validate this column: return the compared against itemsize"""
        ...
    
    def validate_attr(self, append: bool) -> None:
        ...
    
    def update_info(self, info) -> None:
        """
        set/update the info for this indexable with the key/value
        if there is a conflict raise/warn as needed
        """
        ...
    
    def set_info(self, info) -> None:
        """set my state from the passed info"""
        ...
    
    def set_attr(self) -> None:
        """set the kind for this column"""
        ...
    
    def validate_metadata(self, handler: AppendableTable) -> None:
        """validate that kind=category does not change the categories"""
        ...
    
    def write_metadata(self, handler: AppendableTable) -> None:
        """set the meta data"""
        ...
    


class GenericIndexCol(IndexCol):
    """an index which is not represented in the data of the table"""
    @property
    def is_indexed(self) -> bool:
        ...
    
    def convert(self, values: np.ndarray, nan_rep, encoding: str, errors: str) -> tuple[Int64Index, Int64Index]:
        """
        Convert the data from this selection to the appropriate pandas type.

        Parameters
        ----------
        values : np.ndarray
        nan_rep : str
        encoding : str
        errors : str
        """
        ...
    
    def set_attr(self) -> None:
        ...
    


class DataCol(IndexCol):
    """
    a data holding column, by definition this is not indexable

    Parameters
    ----------
    data   : the actual data
    cname  : the column name in the table to hold the data (typically
                values)
    meta   : a string description of the metadata
    metadata : the actual metadata
    """
    is_an_indexable = ...
    is_data_indexable = ...
    _info_fields = ...
    def __init__(self, name: str, values=..., kind=..., typ=..., cname=..., pos=..., tz=..., ordered=..., table=..., meta=..., metadata=..., dtype: DtypeArg | None = ..., data=...) -> None:
        ...
    
    @property
    def dtype_attr(self) -> str:
        ...
    
    @property
    def meta_attr(self) -> str:
        ...
    
    def __repr__(self) -> str:
        ...
    
    def __eq__(self, other: Any) -> bool:
        """compare 2 col items"""
        ...
    
    def set_data(self, data: ArrayLike) -> None:
        ...
    
    def take_data(self): # -> NDArray[Any] | None:
        """return the data"""
        ...
    
    @classmethod
    def get_atom_string(cls, shape, itemsize):
        ...
    
    @classmethod
    def get_atom_coltype(cls, kind: str) -> type[Col]:
        """return the PyTables column class for this column"""
        ...
    
    @classmethod
    def get_atom_data(cls, shape, kind: str) -> Col:
        ...
    
    @classmethod
    def get_atom_datetime64(cls, shape):
        ...
    
    @classmethod
    def get_atom_timedelta64(cls, shape):
        ...
    
    @property
    def shape(self): # -> Any | None:
        ...
    
    @property
    def cvalues(self): # -> NDArray[Any] | None:
        """return my cython values"""
        ...
    
    def validate_attr(self, append) -> None:
        """validate that we have the same order as the existing & same dtype"""
        ...
    
    def convert(self, values: np.ndarray, nan_rep, encoding: str, errors: str): # -> tuple[Unknown | None, ndarray[Unknown, Unknown] | NDArray[Any] | Categorical | Unknown]:
        """
        Convert the data from this selection to the appropriate pandas type.

        Parameters
        ----------
        values : np.ndarray
        nan_rep :
        encoding : str
        errors : str

        Returns
        -------
        index : listlike to become an Index
        data : ndarraylike to become a column
        """
        ...
    
    def set_attr(self) -> None:
        """set the data for this column"""
        ...
    


class DataIndexableCol(DataCol):
    """represent a data column that can be indexed"""
    is_data_indexable = ...
    def validate_names(self) -> None:
        ...
    
    @classmethod
    def get_atom_string(cls, shape, itemsize):
        ...
    
    @classmethod
    def get_atom_data(cls, shape, kind: str) -> Col:
        ...
    
    @classmethod
    def get_atom_datetime64(cls, shape):
        ...
    
    @classmethod
    def get_atom_timedelta64(cls, shape):
        ...
    


class GenericDataIndexableCol(DataIndexableCol):
    """represent a generic pytables data column"""
    ...


class Fixed:
    """
    represent an object in my store
    facilitate read/write of various types of objects
    this is an abstract base class

    Parameters
    ----------
    parent : HDFStore
    group : Node
        The group node where the table resides.
    """
    pandas_kind: str
    format_type: str = ...
    obj_type: type[DataFrame | Series]
    ndim: int
    encoding: str
    parent: HDFStore
    group: Node
    errors: str
    is_table: bool = ...
    def __init__(self, parent: HDFStore, group: Node, encoding: str = ..., errors: str = ...) -> None:
        ...
    
    @property
    def is_old_version(self) -> bool:
        ...
    
    @property
    def version(self) -> tuple[int, int, int]:
        """compute and set our version"""
        ...
    
    @property
    def pandas_type(self): # -> str | Any | None:
        ...
    
    def __repr__(self) -> str:
        """return a pretty representation of myself"""
        ...
    
    def set_object_info(self) -> None:
        """set my pandas type & version"""
        ...
    
    def copy(self) -> Fixed:
        ...
    
    @property
    def shape(self): # -> Any | None:
        ...
    
    @property
    def pathname(self):
        ...
    
    @property
    def attrs(self):
        ...
    
    def set_attrs(self) -> None:
        """set our object attributes"""
        ...
    
    def get_attrs(self) -> None:
        """get our object attributes"""
        ...
    
    @property
    def storable(self):
        """return my storable"""
        ...
    
    @property
    def is_exists(self) -> bool:
        ...
    
    @property
    def nrows(self): # -> Any | None:
        ...
    
    def validate(self, other) -> Literal[True] | None:
        """validate against an existing storable"""
        ...
    
    def validate_version(self, where=...) -> None:
        """are we trying to operate on an old version?"""
        ...
    
    def infer_axes(self) -> bool:
        """
        infer the axes of my storer
        return a boolean indicating if we have a valid storer or not
        """
        ...
    
    def read(self, where=..., columns=..., start: int | None = ..., stop: int | None = ...):
        ...
    
    def write(self, **kwargs):
        ...
    
    def delete(self, where=..., start: int | None = ..., stop: int | None = ...) -> None:
        """
        support fully deleting the node in its entirety (only) - where
        specification must be None
        """
        ...
    


class GenericFixed(Fixed):
    """a generified fixed version"""
    _index_type_map = ...
    _reverse_index_map = ...
    attributes: list[str] = ...
    def validate_read(self, columns, where) -> None:
        """
        raise if any keywords are passed which are not-None
        """
        ...
    
    @property
    def is_exists(self) -> bool:
        ...
    
    def set_attrs(self) -> None:
        """set our object attributes"""
        ...
    
    def get_attrs(self) -> None:
        """retrieve our attributes"""
        ...
    
    def write(self, obj, **kwargs) -> None:
        ...
    
    def read_array(self, key: str, start: int | None = ..., stop: int | None = ...): # -> Any | ndarray[Unknown, Unknown] | NDArray[Any]:
        """read an array for the specified node (off of group"""
        ...
    
    def read_index(self, key: str, start: int | None = ..., stop: int | None = ...) -> Index:
        ...
    
    def write_index(self, key: str, index: Index) -> None:
        ...
    
    def write_multi_index(self, key: str, index: MultiIndex) -> None:
        ...
    
    def read_multi_index(self, key: str, start: int | None = ..., stop: int | None = ...) -> MultiIndex:
        ...
    
    def read_index_node(self, node: Node, start: int | None = ..., stop: int | None = ...) -> Index:
        ...
    
    def write_array_empty(self, key: str, value: ArrayLike) -> None:
        """write a 0-len array"""
        ...
    
    def write_array(self, key: str, obj: AnyArrayLike, items: Index | None = ...) -> None:
        ...
    


class SeriesFixed(GenericFixed):
    pandas_kind = ...
    attributes = ...
    name: Hashable
    @property
    def shape(self): # -> tuple[int] | None:
        ...
    
    def read(self, where=..., columns=..., start: int | None = ..., stop: int | None = ...) -> Series:
        ...
    
    def write(self, obj, **kwargs) -> None:
        ...
    


class BlockManagerFixed(GenericFixed):
    attributes = ...
    nblocks: int
    @property
    def shape(self) -> Shape | None:
        ...
    
    def read(self, where=..., columns=..., start: int | None = ..., stop: int | None = ...) -> DataFrame:
        ...
    
    def write(self, obj, **kwargs) -> None:
        ...
    


class FrameFixed(BlockManagerFixed):
    pandas_kind = ...
    obj_type = DataFrame


class Table(Fixed):
    """
    represent a table:
        facilitate read/write of various types of tables

    Attrs in Table Node
    -------------------
    These are attributes that are store in the main table node, they are
    necessary to recreate these tables when read back in.

    index_axes    : a list of tuples of the (original indexing axis and
        index column)
    non_index_axes: a list of tuples of the (original index axis and
        columns on a non-indexing axis)
    values_axes   : a list of the columns which comprise the data of this
        table
    data_columns  : a list of the columns that we are allowing indexing
        (these become single columns in values_axes)
    nan_rep       : the string to use for nan representations for string
        objects
    levels        : the names of levels
    metadata      : the names of the metadata columns
    """
    pandas_kind = ...
    format_type: str = ...
    table_type: str
    levels: int | list[Hashable] = ...
    is_table = ...
    index_axes: list[IndexCol]
    non_index_axes: list[tuple[int, Any]]
    values_axes: list[DataCol]
    data_columns: list
    metadata: list
    info: dict
    def __init__(self, parent: HDFStore, group: Node, encoding=..., errors: str = ..., index_axes=..., non_index_axes=..., values_axes=..., data_columns=..., info=..., nan_rep=...) -> None:
        ...
    
    @property
    def table_type_short(self) -> str:
        ...
    
    def __repr__(self) -> str:
        """return a pretty representation of myself"""
        ...
    
    def __getitem__(self, c: str): # -> IndexCol | None:
        """return the axis for c"""
        ...
    
    def validate(self, other) -> None:
        """validate against an existing table"""
        ...
    
    @property
    def is_multi_index(self) -> bool:
        """the levels attribute is 1 or a list in the case of a multi-index"""
        ...
    
    def validate_multiindex(self, obj: DataFrame | Series) -> tuple[DataFrame, list[Hashable]]:
        """
        validate that we can store the multi-index; reset and return the
        new object
        """
        ...
    
    @property
    def nrows_expected(self) -> int:
        """based on our axes, compute the expected nrows"""
        ...
    
    @property
    def is_exists(self) -> bool:
        """has this table been created"""
        ...
    
    @property
    def storable(self): # -> Any | None:
        ...
    
    @property
    def table(self): # -> Any | None:
        """return the table group (this is my storable)"""
        ...
    
    @property
    def dtype(self): # -> Any:
        ...
    
    @property
    def description(self): # -> Any:
        ...
    
    @property
    def axes(self): # -> chain[IndexCol]:
        ...
    
    @property
    def ncols(self) -> int:
        """the number of total columns in the values axes"""
        ...
    
    @property
    def is_transposed(self) -> bool:
        ...
    
    @property
    def data_orientation(self) -> tuple[int, ...]:
        """return a tuple of my permutated axes, non_indexable at the front"""
        ...
    
    def queryables(self) -> dict[str, Any]:
        """return a dict of the kinds allowable columns for this object"""
        ...
    
    def index_cols(self): # -> list[tuple[Unknown | None, str]]:
        """return a list of my index cols"""
        ...
    
    def values_cols(self) -> list[str]:
        """return a list of my values cols"""
        ...
    
    def write_metadata(self, key: str, values: np.ndarray) -> None:
        """
        Write out a metadata array to the key as a fixed-format Series.

        Parameters
        ----------
        key : str
        values : ndarray
        """
        ...
    
    def read_metadata(self, key: str): # -> TableIterator | None:
        """return the meta data array for this key"""
        ...
    
    def set_attrs(self) -> None:
        """set our table type & indexables"""
        ...
    
    def get_attrs(self) -> None:
        """retrieve our attributes"""
        ...
    
    def validate_version(self, where=...) -> None:
        """are we trying to operate on an old version?"""
        ...
    
    def validate_min_itemsize(self, min_itemsize) -> None:
        """
        validate the min_itemsize doesn't contain items that are not in the
        axes this needs data_columns to be defined
        """
        ...
    
    @cache_readonly
    def indexables(self): # -> list[Unknown]:
        """create/cache the indexables if they don't exist"""
        ...
    
    def create_index(self, columns=..., optlevel=..., kind: str | None = ...) -> None:
        """
        Create a pytables index on the specified columns.

        Parameters
        ----------
        columns : None, bool, or listlike[str]
            Indicate which columns to create an index on.

            * False : Do not create any indexes.
            * True : Create indexes on all columns.
            * None : Create indexes on all columns.
            * listlike : Create indexes on the given columns.

        optlevel : int or None, default None
            Optimization level, if None, pytables defaults to 6.
        kind : str or None, default None
            Kind of index, if None, pytables defaults to "medium".

        Raises
        ------
        TypeError if trying to create an index on a complex-type column.

        Notes
        -----
        Cannot index Time64Col or ComplexCol.
        Pytables must be >= 3.0.
        """
        ...
    
    @classmethod
    def get_object(cls, obj, transposed: bool):
        """return the data for this obj"""
        ...
    
    def validate_data_columns(self, data_columns, min_itemsize, non_index_axes): # -> list[Unknown]:
        """
        take the input data_columns and min_itemize and create a data
        columns spec
        """
        ...
    
    def process_axes(self, obj, selection: Selection, columns=...) -> DataFrame:
        """process axes filters"""
        ...
    
    def create_description(self, complib, complevel: int | None, fletcher32: bool, expectedrows: int | None) -> dict[str, Any]:
        """create the description of the table from the axes & values"""
        ...
    
    def read_coordinates(self, where=..., start: int | None = ..., stop: int | None = ...): # -> Index | Literal[False]:
        """
        select coordinates (row numbers) from a table; return the
        coordinates object
        """
        ...
    
    def read_column(self, column: str, where=..., start: int | None = ..., stop: int | None = ...): # -> Series | Literal[False]:
        """
        return a single column from the table, generally only indexables
        are interesting
        """
        ...
    


class WORMTable(Table):
    """
    a write-once read-many table: this format DOES NOT ALLOW appending to a
    table. writing is a one-time operation the data are stored in a format
    that allows for searching the data on disk
    """
    table_type = ...
    def read(self, where=..., columns=..., start: int | None = ..., stop: int | None = ...):
        """
        read the indices and the indexing array, calculate offset rows and return
        """
        ...
    
    def write(self, **kwargs) -> None:
        """
        write in a format that we can search later on (but cannot append
        to): write out the indices and the values using _write_array
        (e.g. a CArray) create an indexing table so that we can search
        """
        ...
    


class AppendableTable(Table):
    """support the new appendable table formats"""
    table_type = ...
    def write(self, obj, axes=..., append: bool = ..., complib=..., complevel=..., fletcher32=..., min_itemsize=..., chunksize=..., expectedrows=..., dropna: bool = ..., nan_rep=..., data_columns=..., track_times=...) -> None:
        ...
    
    def write_data(self, chunksize: int | None, dropna: bool = ...) -> None:
        """
        we form the data into a 2-d including indexes,values,mask write chunk-by-chunk
        """
        ...
    
    def write_data_chunk(self, rows: np.ndarray, indexes: list[np.ndarray], mask: npt.NDArray[np.bool_] | None, values: list[np.ndarray]) -> None:
        """
        Parameters
        ----------
        rows : an empty memory space where we are putting the chunk
        indexes : an array of the indexes
        mask : an array of the masks
        values : an array of the values
        """
        ...
    
    def delete(self, where=..., start: int | None = ..., stop: int | None = ...): # -> Any | int | None:
        ...
    


class AppendableFrameTable(AppendableTable):
    """support the new appendable table formats"""
    pandas_kind = ...
    table_type = ...
    ndim = ...
    obj_type: type[DataFrame | Series] = ...
    @property
    def is_transposed(self) -> bool:
        ...
    
    @classmethod
    def get_object(cls, obj, transposed: bool):
        """these are written transposed"""
        ...
    
    def read(self, where=..., columns=..., start: int | None = ..., stop: int | None = ...): # -> DataFrame | None:
        ...
    


class AppendableSeriesTable(AppendableFrameTable):
    """support the new appendable table formats"""
    pandas_kind = ...
    table_type = ...
    ndim = ...
    obj_type = Series
    @property
    def is_transposed(self) -> bool:
        ...
    
    @classmethod
    def get_object(cls, obj, transposed: bool):
        ...
    
    def write(self, obj, data_columns=..., **kwargs): # -> None:
        """we are going to write this as a frame table"""
        ...
    
    def read(self, where=..., columns=..., start: int | None = ..., stop: int | None = ...) -> Series:
        ...
    


class AppendableMultiSeriesTable(AppendableSeriesTable):
    """support the new appendable table formats"""
    pandas_kind = ...
    table_type = ...
    def write(self, obj, **kwargs): # -> None:
        """we are going to write this as a frame table"""
        ...
    


class GenericTable(AppendableFrameTable):
    """a table that read/writes the generic pytables table format"""
    pandas_kind = ...
    table_type = ...
    ndim = ...
    obj_type = DataFrame
    levels: list[Hashable]
    @property
    def pandas_type(self) -> str:
        ...
    
    @property
    def storable(self): # -> Any:
        ...
    
    def get_attrs(self) -> None:
        """retrieve our attributes"""
        ...
    
    @cache_readonly
    def indexables(self): # -> list[GenericIndexCol | GenericDataIndexableCol]:
        """create the indexables from the table description"""
        ...
    
    def write(self, **kwargs):
        ...
    


class AppendableMultiFrameTable(AppendableFrameTable):
    """a frame with a multi-index"""
    table_type = ...
    obj_type = DataFrame
    ndim = ...
    _re_levels = ...
    @property
    def table_type_short(self) -> str:
        ...
    
    def write(self, obj, data_columns=..., **kwargs): # -> None:
        ...
    
    def read(self, where=..., columns=..., start: int | None = ..., stop: int | None = ...): # -> DataFrame:
        ...
    


class Selection:
    """
    Carries out a selection operation on a tables.Table object.

    Parameters
    ----------
    table : a Table object
    where : list of Terms (or convertible to)
    start, stop: indices to start and/or stop selection

    """
    def __init__(self, table: Table, where=..., start: int | None = ..., stop: int | None = ...) -> None:
        ...
    
    def generate(self, where): # -> PyTablesExpr | None:
        """where can be a : dict,list,tuple,string"""
        ...
    
    def select(self): # -> Any:
        """
        generate the selection
        """
        ...
    
    def select_coords(self): # -> Any | NDArray[Any] | NDArray[Unknown]:
        """
        generate the selection
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from abc import ABC, abstractmethod
from typing import Any, Callable, Hashable, Sequence, TYPE_CHECKING
from pandas._typing import NDFrameT
from pandas import DataFrame, Series

"""
Module responsible for execution of NDFrame.describe() method.

Method NDFrame.describe() delegates actual execution to function describe_ndframe().
"""
if TYPE_CHECKING:
    ...
def describe_ndframe(*, obj: NDFrameT, include: str | Sequence[str] | None, exclude: str | Sequence[str] | None, datetime_is_numeric: bool, percentiles: Sequence[float] | np.ndarray | None) -> NDFrameT:
    """Describe series or dataframe.

    Called from pandas.core.generic.NDFrame.describe()

    Parameters
    ----------
    obj: DataFrame or Series
        Either dataframe or series to be described.
    include : 'all', list-like of dtypes or None (default), optional
        A white list of data types to include in the result. Ignored for ``Series``.
    exclude : list-like of dtypes or None (default), optional,
        A black list of data types to omit from the result. Ignored for ``Series``.
    datetime_is_numeric : bool, default False
        Whether to treat datetime dtypes as numeric.
    percentiles : list-like of numbers, optional
        The percentiles to include in the output. All should fall between 0 and 1.
        The default is ``[.25, .5, .75]``, which returns the 25th, 50th, and
        75th percentiles.

    Returns
    -------
    Dataframe or series description.
    """
    ...

class NDFrameDescriberAbstract(ABC):
    """Abstract class for describing dataframe or series.

    Parameters
    ----------
    obj : Series or DataFrame
        Object to be described.
    datetime_is_numeric : bool
        Whether to treat datetime dtypes as numeric.
    """
    def __init__(self, obj: DataFrame | Series, datetime_is_numeric: bool) -> None:
        ...
    
    @abstractmethod
    def describe(self, percentiles: Sequence[float] | np.ndarray) -> DataFrame | Series:
        """Do describe either series or dataframe.

        Parameters
        ----------
        percentiles : list-like of numbers
            The percentiles to include in the output.
        """
        ...
    


class SeriesDescriber(NDFrameDescriberAbstract):
    """Class responsible for creating series description."""
    obj: Series
    def describe(self, percentiles: Sequence[float] | np.ndarray) -> Series:
        ...
    


class DataFrameDescriber(NDFrameDescriberAbstract):
    """Class responsible for creating dataobj description.

    Parameters
    ----------
    obj : DataFrame
        DataFrame to be described.
    include : 'all', list-like of dtypes or None
        A white list of data types to include in the result.
    exclude : list-like of dtypes or None
        A black list of data types to omit from the result.
    datetime_is_numeric : bool
        Whether to treat datetime dtypes as numeric.
    """
    def __init__(self, obj: DataFrame, *, include: str | Sequence[str] | None, exclude: str | Sequence[str] | None, datetime_is_numeric: bool) -> None:
        ...
    
    def describe(self, percentiles: Sequence[float] | np.ndarray) -> DataFrame:
        ...
    


def reorder_columns(ldesc: Sequence[Series]) -> list[Hashable]:
    """Set a convenient order for rows for display."""
    ...

def describe_numeric_1d(series: Series, percentiles: Sequence[float]) -> Series:
    """Describe series containing numerical data.

    Parameters
    ----------
    series : Series
        Series to be described.
    percentiles : list-like of numbers
        The percentiles to include in the output.
    """
    ...

def describe_categorical_1d(data: Series, percentiles_ignored: Sequence[float]) -> Series:
    """Describe series containing categorical data.

    Parameters
    ----------
    data : Series
        Series to be described.
    percentiles_ignored : list-like of numbers
        Ignored, but in place to unify interface.
    """
    ...

def describe_timestamp_as_categorical_1d(data: Series, percentiles_ignored: Sequence[float]) -> Series:
    """Describe series containing timestamp data treated as categorical.

    Parameters
    ----------
    data : Series
        Series to be described.
    percentiles_ignored : list-like of numbers
        Ignored, but in place to unify interface.
    """
    ...

def describe_timestamp_1d(data: Series, percentiles: Sequence[float]) -> Series:
    """Describe series containing datetime64 dtype.

    Parameters
    ----------
    data : Series
        Series to be described.
    percentiles : list-like of numbers
        The percentiles to include in the output.
    """
    ...

def select_describe_func(data: Series, datetime_is_numeric: bool) -> Callable:
    """Select proper function for describing series based on data type.

    Parameters
    ----------
    data : Series
        Series to be described.
    datetime_is_numeric : bool
        Whether to treat datetime dtypes as numeric.
    """
    ...

def refine_percentiles(percentiles: Sequence[float] | np.ndarray | None) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Ensure that percentiles are unique and sorted.

    Parameters
    ----------
    percentiles : list-like of numbers, optional
        The percentiles to include in the output.
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import TYPE_CHECKING, overload
from pandas._typing import ArrayLike, DtypeObj, IgnoreRaise
from pandas.core.dtypes.dtypes import ExtensionDtype
from pandas.core.arrays import DatetimeArray, ExtensionArray

"""
Functions for implementing 'astype' methods according to pandas conventions,
particularly ones that differ from numpy.
"""
if TYPE_CHECKING:
    ...
_dtype_obj = ...
@overload
def astype_nansafe(arr: np.ndarray, dtype: np.dtype, copy: bool = ..., skipna: bool = ...) -> np.ndarray:
    ...

@overload
def astype_nansafe(arr: np.ndarray, dtype: ExtensionDtype, copy: bool = ..., skipna: bool = ...) -> ExtensionArray:
    ...

def astype_nansafe(arr: np.ndarray, dtype: DtypeObj, copy: bool = ..., skipna: bool = ...) -> ArrayLike:
    """
    Cast the elements of an array to a given dtype a nan-safe manner.

    Parameters
    ----------
    arr : ndarray
    dtype : np.dtype or ExtensionDtype
    copy : bool, default True
        If False, a view will be attempted but may fail, if
        e.g. the item sizes don't align.
    skipna: bool, default False
        Whether or not we should skip NaN when casting as a string-type.

    Raises
    ------
    ValueError
        The dtype was a datetime64/timedelta64 dtype, but it had no unit.
    """
    ...

def astype_array(values: ArrayLike, dtype: DtypeObj, copy: bool = ...) -> ArrayLike:
    """
    Cast array (ndarray or ExtensionArray) to the new dtype.

    Parameters
    ----------
    values : ndarray or ExtensionArray
    dtype : dtype object
    copy : bool, default False
        copy if indicated

    Returns
    -------
    ndarray or ExtensionArray
    """
    ...

def astype_array_safe(values: ArrayLike, dtype, copy: bool = ..., errors: IgnoreRaise = ...) -> ArrayLike:
    """
    Cast array (ndarray or ExtensionArray) to the new dtype.

    This basically is the implementation for DataFrame/Series.astype and
    includes all custom logic for pandas (NaN-safety, converting str to object,
    not allowing )

    Parameters
    ----------
    values : ndarray or ExtensionArray
    dtype : str, dtype convertible
    copy : bool, default False
        copy if indicated
    errors : str, {'raise', 'ignore'}, default 'raise'
        - ``raise`` : allow exceptions to be raised
        - ``ignore`` : suppress exceptions. On error return original object

    Returns
    -------
    ndarray or ExtensionArray
    """
    ...

def astype_td64_unit_conversion(values: np.ndarray, dtype: np.dtype, copy: bool) -> np.ndarray:
    """
    By pandas convention, converting to non-nano timedelta64
    returns an int64-dtyped array with ints representing multiples
    of the desired timedelta unit.  This is essentially division.

    Parameters
    ----------
    values : np.ndarray[timedelta64[ns]]
    dtype : np.dtype
        timedelta64 with unit not-necessarily nano
    copy : bool

    Returns
    -------
    np.ndarray
    """
    ...

def astype_dt64_to_dt64tz(values: ArrayLike, dtype: DtypeObj, copy: bool, via_utc: bool = ...) -> DatetimeArray:
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from datetime import datetime, timedelta
from typing import Any, Collection, Literal, TYPE_CHECKING, TypeVar, overload
from pandas._typing import npt
from .timedeltas import Timedelta
from pandas.core.indexes.datetimes import DatetimeIndex

if TYPE_CHECKING:
    ...
_BaseOffsetT = TypeVar("_BaseOffsetT", bound=BaseOffset)
_DatetimeT = TypeVar("_DatetimeT", bound=datetime)
_TimedeltaT = TypeVar("_TimedeltaT", bound=timedelta)
_relativedelta_kwds: set[str]
prefix_mapping: dict[str, type]
class ApplyTypeError(TypeError):
    ...


class BaseOffset:
    n: int
    def __init__(self, n: int = ..., normalize: bool = ...) -> None:
        ...
    
    def __eq__(self, other) -> bool:
        ...
    
    def __ne__(self, other) -> bool:
        ...
    
    def __hash__(self) -> int:
        ...
    
    @property
    def kwds(self) -> dict:
        ...
    
    @property
    def base(self) -> BaseOffset:
        ...
    
    @overload
    def __add__(self, other: npt.NDArray[np.object_]) -> npt.NDArray[np.object_]:
        ...
    
    @overload
    def __add__(self: _BaseOffsetT, other: BaseOffset) -> _BaseOffsetT:
        ...
    
    @overload
    def __add__(self, other: _DatetimeT) -> _DatetimeT:
        ...
    
    @overload
    def __add__(self, other: _TimedeltaT) -> _TimedeltaT:
        ...
    
    @overload
    def __radd__(self, other: npt.NDArray[np.object_]) -> npt.NDArray[np.object_]:
        ...
    
    @overload
    def __radd__(self: _BaseOffsetT, other: BaseOffset) -> _BaseOffsetT:
        ...
    
    @overload
    def __radd__(self, other: _DatetimeT) -> _DatetimeT:
        ...
    
    @overload
    def __radd__(self, other: _TimedeltaT) -> _TimedeltaT:
        ...
    
    def __sub__(self: _BaseOffsetT, other: BaseOffset) -> _BaseOffsetT:
        ...
    
    @overload
    def __rsub__(self, other: npt.NDArray[np.object_]) -> npt.NDArray[np.object_]:
        ...
    
    @overload
    def __rsub__(self: _BaseOffsetT, other: BaseOffset) -> _BaseOffsetT:
        ...
    
    @overload
    def __rsub__(self, other: _DatetimeT) -> _DatetimeT:
        ...
    
    @overload
    def __rsub__(self, other: _TimedeltaT) -> _TimedeltaT:
        ...
    
    def __call__(self, other):
        ...
    
    @overload
    def __mul__(self, other: np.ndarray) -> np.ndarray:
        ...
    
    @overload
    def __mul__(self: _BaseOffsetT, other: int) -> _BaseOffsetT:
        ...
    
    @overload
    def __rmul__(self, other: np.ndarray) -> np.ndarray:
        ...
    
    @overload
    def __rmul__(self: _BaseOffsetT, other: int) -> _BaseOffsetT:
        ...
    
    def __neg__(self: _BaseOffsetT) -> _BaseOffsetT:
        ...
    
    def copy(self: _BaseOffsetT) -> _BaseOffsetT:
        ...
    
    @property
    def name(self) -> str:
        ...
    
    @property
    def rule_code(self) -> str:
        ...
    
    @property
    def freqstr(self) -> str:
        ...
    
    def apply_index(self, dtindex: DatetimeIndex) -> DatetimeIndex:
        ...
    
    def rollback(self, dt: datetime) -> datetime:
        ...
    
    def rollforward(self, dt: datetime) -> datetime:
        ...
    
    def is_on_offset(self, dt: datetime) -> bool:
        ...
    
    def __setstate__(self, state) -> None:
        ...
    
    def __getstate__(self):
        ...
    
    @property
    def nanos(self) -> int:
        ...
    
    def onOffset(self, dt: datetime) -> bool:
        ...
    
    def isAnchored(self) -> bool:
        ...
    
    def is_anchored(self) -> bool:
        ...
    


class SingleConstructorOffset(BaseOffset):
    def __reduce__(self):
        ...
    


@overload
def to_offset(freq: None) -> None:
    ...

@overload
def to_offset(freq: _BaseOffsetT) -> _BaseOffsetT:
    ...

@overload
def to_offset(freq: timedelta | str) -> BaseOffset:
    ...

class Tick(SingleConstructorOffset):
    _reso: int
    _prefix: str
    _td64_unit: str
    def __init__(self, n: int = ..., normalize: bool = ...) -> None:
        ...
    
    @property
    def delta(self) -> Timedelta:
        ...
    
    @property
    def nanos(self) -> int:
        ...
    


def delta_to_tick(delta: timedelta) -> Tick:
    ...

class Day(Tick):
    ...


class Hour(Tick):
    ...


class Minute(Tick):
    ...


class Second(Tick):
    ...


class Milli(Tick):
    ...


class Micro(Tick):
    ...


class Nano(Tick):
    ...


class RelativeDeltaOffset(BaseOffset):
    def __init__(self, n: int = ..., normalize: bool = ..., **kwds: Any) -> None:
        ...
    


class BusinessMixin(SingleConstructorOffset):
    def __init__(self, n: int = ..., normalize: bool = ..., offset: timedelta = ...) -> None:
        ...
    


class BusinessDay(BusinessMixin):
    ...


class BusinessHour(BusinessMixin):
    def __init__(self, n: int = ..., normalize: bool = ..., start: str | Collection[str] = ..., end: str | Collection[str] = ..., offset: timedelta = ...) -> None:
        ...
    


class WeekOfMonthMixin(SingleConstructorOffset):
    def __init__(self, n: int = ..., normalize: bool = ..., weekday: int = ...) -> None:
        ...
    


class YearOffset(SingleConstructorOffset):
    def __init__(self, n: int = ..., normalize: bool = ..., month: int | None = ...) -> None:
        ...
    


class BYearEnd(YearOffset):
    ...


class BYearBegin(YearOffset):
    ...


class YearEnd(YearOffset):
    ...


class YearBegin(YearOffset):
    ...


class QuarterOffset(SingleConstructorOffset):
    def __init__(self, n: int = ..., normalize: bool = ..., startingMonth: int | None = ...) -> None:
        ...
    


class BQuarterEnd(QuarterOffset):
    ...


class BQuarterBegin(QuarterOffset):
    ...


class QuarterEnd(QuarterOffset):
    ...


class QuarterBegin(QuarterOffset):
    ...


class MonthOffset(SingleConstructorOffset):
    ...


class MonthEnd(MonthOffset):
    ...


class MonthBegin(MonthOffset):
    ...


class BusinessMonthEnd(MonthOffset):
    ...


class BusinessMonthBegin(MonthOffset):
    ...


class SemiMonthOffset(SingleConstructorOffset):
    def __init__(self, n: int = ..., normalize: bool = ..., day_of_month: int | None = ...) -> None:
        ...
    


class SemiMonthEnd(SemiMonthOffset):
    ...


class SemiMonthBegin(SemiMonthOffset):
    ...


class Week(SingleConstructorOffset):
    def __init__(self, n: int = ..., normalize: bool = ..., weekday: int | None = ...) -> None:
        ...
    


class WeekOfMonth(WeekOfMonthMixin):
    def __init__(self, n: int = ..., normalize: bool = ..., week: int = ..., weekday: int = ...) -> None:
        ...
    


class LastWeekOfMonth(WeekOfMonthMixin):
    ...


class FY5253Mixin(SingleConstructorOffset):
    def __init__(self, n: int = ..., normalize: bool = ..., weekday: int = ..., startingMonth: int = ..., variation: Literal["nearest", "last"] = ...) -> None:
        ...
    


class FY5253(FY5253Mixin):
    ...


class FY5253Quarter(FY5253Mixin):
    def __init__(self, n: int = ..., normalize: bool = ..., weekday: int = ..., startingMonth: int = ..., qtr_with_extra_week: int = ..., variation: Literal["nearest", "last"] = ...) -> None:
        ...
    


class Easter(SingleConstructorOffset):
    ...


class _CustomBusinessMonth(BusinessMixin):
    def __init__(self, n: int = ..., normalize: bool = ..., weekmask: str = ..., holidays: list | None = ..., calendar: np.busdaycalendar | None = ..., offset: timedelta = ...) -> None:
        ...
    


class CustomBusinessDay(BusinessDay):
    def __init__(self, n: int = ..., normalize: bool = ..., weekmask: str = ..., holidays: list | None = ..., calendar: np.busdaycalendar | None = ..., offset: timedelta = ...) -> None:
        ...
    


class CustomBusinessHour(BusinessHour):
    def __init__(self, n: int = ..., normalize: bool = ..., weekmask: str = ..., holidays: list | None = ..., calendar: np.busdaycalendar | None = ..., start: str = ..., end: str = ..., offset: timedelta = ...) -> None:
        ...
    


class CustomBusinessMonthEnd(_CustomBusinessMonth):
    ...


class CustomBusinessMonthBegin(_CustomBusinessMonth):
    ...


class DateOffset(RelativeDeltaOffset):
    ...


BDay = BusinessDay
BMonthEnd = BusinessMonthEnd
BMonthBegin = BusinessMonthBegin
CBMonthEnd = CustomBusinessMonthEnd
CBMonthBegin = CustomBusinessMonthBegin
CDay = CustomBusinessDay
def roll_qtrday(other: datetime, n: int, month: int, day_opt: str, modby: int) -> int:
    ...

INVALID_FREQ_ERR_MSG: Literal["Invalid frequency: {0}"]
def shift_months(dtindex: npt.NDArray[np.int64], months: int, day_opt: str | None = ...) -> npt.NDArray[np.int64]:
    ...

_offset_map: dict[str, BaseOffset]



"""
This type stub file was generated by pyright.
"""

from datetime import timedelta
from typing import TYPE_CHECKING, overload
from pandas._libs.tslibs.timedeltas import Timedelta, UnitChoices
from pandas._typing import ArrayLike, DateTimeErrorChoices
from pandas import Index, Series, TimedeltaIndex

"""
timedelta support tools
"""
if TYPE_CHECKING:
    ...
@overload
def to_timedelta(arg: str | float | timedelta, unit: UnitChoices | None = ..., errors: DateTimeErrorChoices = ...) -> Timedelta:
    ...

@overload
def to_timedelta(arg: Series, unit: UnitChoices | None = ..., errors: DateTimeErrorChoices = ...) -> Series:
    ...

@overload
def to_timedelta(arg: list | tuple | range | ArrayLike | Index, unit: UnitChoices | None = ..., errors: DateTimeErrorChoices = ...) -> TimedeltaIndex:
    ...

def to_timedelta(arg: str | int | float | timedelta | list | tuple | range | ArrayLike | Index | Series, unit: UnitChoices | None = ..., errors: DateTimeErrorChoices = ...) -> Timedelta | TimedeltaIndex | Series:
    """
    Convert argument to timedelta.

    Timedeltas are absolute differences in times, expressed in difference
    units (e.g. days, hours, minutes, seconds). This method converts
    an argument from a recognized timedelta format / value into
    a Timedelta type.

    Parameters
    ----------
    arg : str, timedelta, list-like or Series
        The data to be converted to timedelta.

        .. deprecated:: 1.2
            Strings with units 'M', 'Y' and 'y' do not represent
            unambiguous timedelta values and will be removed in a future version

    unit : str, optional
        Denotes the unit of the arg for numeric `arg`. Defaults to ``"ns"``.

        Possible values:

        * 'W'
        * 'D' / 'days' / 'day'
        * 'hours' / 'hour' / 'hr' / 'h'
        * 'm' / 'minute' / 'min' / 'minutes' / 'T'
        * 'S' / 'seconds' / 'sec' / 'second'
        * 'ms' / 'milliseconds' / 'millisecond' / 'milli' / 'millis' / 'L'
        * 'us' / 'microseconds' / 'microsecond' / 'micro' / 'micros' / 'U'
        * 'ns' / 'nanoseconds' / 'nano' / 'nanos' / 'nanosecond' / 'N'

        .. versionchanged:: 1.1.0

           Must not be specified when `arg` context strings and
           ``errors="raise"``.

    errors : {'ignore', 'raise', 'coerce'}, default 'raise'
        - If 'raise', then invalid parsing will raise an exception.
        - If 'coerce', then invalid parsing will be set as NaT.
        - If 'ignore', then invalid parsing will return the input.

    Returns
    -------
    timedelta
        If parsing succeeded.
        Return type depends on input:

        - list-like: TimedeltaIndex of timedelta64 dtype
        - Series: Series of timedelta64 dtype
        - scalar: Timedelta

    See Also
    --------
    DataFrame.astype : Cast argument to a specified dtype.
    to_datetime : Convert argument to datetime.
    convert_dtypes : Convert dtypes.

    Notes
    -----
    If the precision is higher than nanoseconds, the precision of the duration is
    truncated to nanoseconds for string inputs.

    Examples
    --------
    Parsing a single string to a Timedelta:

    >>> pd.to_timedelta('1 days 06:05:01.00003')
    Timedelta('1 days 06:05:01.000030')
    >>> pd.to_timedelta('15.5us')
    Timedelta('0 days 00:00:00.000015500')

    Parsing a list or array of strings:

    >>> pd.to_timedelta(['1 days 06:05:01.00003', '15.5us', 'nan'])
    TimedeltaIndex(['1 days 06:05:01.000030', '0 days 00:00:00.000015500', NaT],
                   dtype='timedelta64[ns]', freq=None)

    Converting numbers by specifying the `unit` keyword argument:

    >>> pd.to_timedelta(np.arange(5), unit='s')
    TimedeltaIndex(['0 days 00:00:00', '0 days 00:00:01', '0 days 00:00:02',
                    '0 days 00:00:03', '0 days 00:00:04'],
                   dtype='timedelta64[ns]', freq=None)
    >>> pd.to_timedelta(np.arange(5), unit='d')
    TimedeltaIndex(['0 days', '1 days', '2 days', '3 days', '4 days'],
                   dtype='timedelta64[ns]', freq=None)
    """
    ...




"""
This type stub file was generated by pyright.
"""

from pandas.core.arrays.arrow.array import ArrowExtensionArray
from pandas.core.arrays.arrow.dtype import ArrowDtype

__all__ = ["ArrowDtype", "ArrowExtensionArray"]



# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Union
from boario.simulation import Simulation
import numpyencoder
import json
import pathlib
import numpy as np
import pandas as pd
import itertools
from boario.utils import misc
import dask.dataframe as da
from boario import logger

__all__ = ['Indicators']

def df_from_memmap(memmap:np.memmap, indexes:dict) -> pd.DataFrame:
    a = pd.DataFrame(memmap, columns=pd.MultiIndex.from_product([indexes["regions"], indexes["sectors"]]))
    a['step'] = a.index
    return a

def stock_df_from_memmap(memmap:np.memmap, indexes:dict, timesteps:int, timesteps_simulated:int) -> pd.DataFrame:
    a = pd.DataFrame(memmap.reshape(timesteps*indexes["n_sectors"],-1), index=pd.MultiIndex.from_product([timesteps, indexes["sectors"]], names=['step', 'stock of']), columns=pd.MultiIndex.from_product([indexes["regions"], indexes["sectors"]]))
    a = a.loc[pd.IndexSlice[:timesteps_simulated,:]]
    return a

class Indicators(object):

    record_files_list = ["final_demand_record",
                         "io_demand_record",
                         "final_demand_unmet_record",
                         "iotable_X_max_record",
                         "iotable_XVA_record",
                         "limiting_stocks_record",
                         "overprodvector_record",
                         "rebuild_demand_record",
                         "rebuild_prod_record"
                         ]

    params_list = ["simulated_params", "simulated_events"]

    def __init__(self, data_dict:dict, include_crash:bool = False) -> None:
        logger.info("Instanciating indicators")
        super().__init__()
        if not include_crash:
            if data_dict["has_crashed"]:
                raise RuntimeError("Simulation crashed and include_crash is False, I won't compute indicators")
        steps = [i for i in range(data_dict["n_temporal_units_to_sim"])]

        if "stocks" in data_dict:
            stock_treatement = True
        else:
            stock_treatement = False

        self.params = data_dict["params"]
        self.n_rows = data_dict["n_temporal_units_to_sim"]
        self.temporal_units_by_step = data_dict["temporal_units_by_step"]

        self.prod_df = pd.DataFrame(data_dict["prod"], columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region','sector']))
        self.prod_df = self.prod_df.interpolate()
        self.prodmax_df = pd.DataFrame(data_dict["prodmax"], columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region', 'sector']))
        self.prodmax_df = self.prodmax_df.interpolate()
        self.overprod_df = pd.DataFrame(data_dict["overprod"], columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region', 'sector']))
        self.overprod_df = self.overprod_df.interpolate()
        self.final_demand_df = pd.DataFrame(data_dict["final_demand"], columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region', 'sector']))
        self.final_demand_df = self.final_demand_df.interpolate()
        self.io_demand_df = pd.DataFrame(data_dict["io_demand"], columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region', 'sector']))
        self.io_demand_df = self.io_demand_df.interpolate()
        self.r_demand_df = pd.DataFrame(data_dict["r_demand"], columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region', 'sector']))
        self.r_demand_df = self.r_demand_df.interpolate()
        self.r_prod_df = pd.DataFrame(data_dict["r_prod"], columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region', 'sector']))
        self.r_prod_df = self.r_prod_df.interpolate()
        fd_unmet_df = pd.DataFrame(data_dict["fd_unmet"], columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region', 'sector']))
        if stock_treatement:
            stocks_df = pd.DataFrame(data_dict["stocks"].reshape(data_dict["n_temporal_units_to_sim"]*data_dict["n_sectors"],-1),
                                     index=pd.MultiIndex.from_product([steps, data_dict["sectors"]], names=['step', 'stock of']),
                                     columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region', 'sector']))
            stocks_df = stocks_df.loc[pd.IndexSlice[:data_dict["n_temporal_units_simulated"],:]]
        else:
            stocks_df = None
        self.prod_df = self.prod_df.rename_axis('step')
        self.prodmax_df = self.prodmax_df.rename_axis('step')
        self.overprod_df = self.overprod_df.rename_axis('step')
        self.final_demand_df = self.final_demand_df.rename_axis('step')
        self.io_demand_df = self.io_demand_df.rename_axis('step')
        self.r_demand_df = self.r_demand_df.rename_axis('step')
        self.r_prod_df = self.r_prod_df.rename_axis('step')
        self.fd_unmet_df = fd_unmet_df.rename_axis('step')
        self.fd_unmet_df = self.fd_unmet_df.interpolate()

        if stock_treatement:
            stocks_df = stocks_df.replace([np.inf, -np.inf], np.nan).dropna(how='all')
            stocks_df = stocks_df.astype(np.float32)
            stocks_df = stocks_df.groupby('stock of').pct_change().fillna(0).add(1).groupby('stock of').cumprod().sub(1) #type: ignore
            stocks_df = stocks_df.melt(ignore_index=False).rename(columns={'variable_0':'region','variable_1':'sector', 'variable_2':'stock of'})
            stocks_df = stocks_df.reset_index()
            stocks_df['step'] = stocks_df['step'].astype("uint16")
            stocks_df['stock of'] = stocks_df['stock of'].astype("category")
            stocks_df['region'] = stocks_df['region'].astype("category")
            stocks_df['sector'] = stocks_df['sector'].astype("category")
        self.df_stocks = stocks_df
        del stocks_df
        # self.df_stocks = self.df_stocks.interpolate()

        self.df_loss = self.fd_unmet_df.melt(ignore_index=False).rename(columns={'variable_0':'region','variable_1':'fd_cat', 'value':'fdloss'}).reset_index()

        self.df_limiting = pd.DataFrame(data_dict["limiting_stocks"].reshape(data_dict["n_temporal_units_to_sim"]*data_dict["n_sectors"],-1),
                                        index=pd.MultiIndex.from_product([steps, data_dict["sectors"]], names=['step', 'stock of']),
                                        columns=pd.MultiIndex.from_product([data_dict["regions"], data_dict["sectors"]], names=['region', 'sector']))
        self.aff_regions = []
        for e in data_dict["events"]:
            self.aff_regions.append(e['aff_regions'])

        self.aff_regions = list(misc.flatten(self.aff_regions))

        self.aff_sectors = []
        for e in data_dict["events"]:
            self.aff_sectors.append(e['aff_sectors'])
        self.aff_sectors = list(misc.flatten(self.aff_sectors))

        self.rebuilding_sectors = []
        for e in data_dict["events"]:
            reb_sec = e.get('rebuilding_sectors')
            if reb_sec is None:
                self.rebuilding_sectors=[]
            else:
                self.rebuilding_sectors.append(e['rebuilding_sectors'].keys())
                self.rebuilding_sectors = list(misc.flatten(self.rebuilding_sectors))

        # As we dump the class __dict__ we have different names...
        # TODO: Find a better fix than this !
        if 'r_damages' in data_dict['events'][0]:
            data_dict['events'][0]['r_dmg'] = data_dict['events'][0]['r_damages']
        if 'q_damages' in data_dict['events'][0]:
            data_dict['events'][0]['q_dmg'] = data_dict['events'][0]['q_damages']

        if 'r_dmg' in data_dict['events'][0]:
            gdp_dmg_share = data_dict['events'][0]['r_dmg']
        else:
            gdp_dmg_share = -1.
        self.indicators = {
            "region" : self.aff_regions,
            "gdp_dmg_share" : gdp_dmg_share,
            "tot_fd_unmet": "unset",
            "aff_fd_unmet": "unset",
            "rebuild_durations": "unset",
            "shortage_b": False,
            "shortage_date_start": "unset",
            "shortage_date_end": "unset",
            "shortage_date_max": "unset",
            "shortage_ind_max": "unset",
            "shortage_ind_mean": "unset",
            "10_first_shortages_(step,region,sector,stock_of)": "unset",
            "prod_gain_tot": "unset",
            "prod_lost_tot": "unset",
            "prod_gain_unaff": "unset",
            "prod_lost_unaff": "unset",
            "psi" : data_dict['params']['psi_param'],
            "inv_tau" : data_dict['params']['inventory_restoration_tau'],
            "n_temporal_units_to_sim" : data_dict['n_temporal_units_simulated'],
            "has_crashed" : data_dict['has_crashed'],
        }
        self.storage = (pathlib.Path(data_dict['results_storage'])/'indicators.json').resolve()
        self.storage_path = (pathlib.Path(data_dict['results_storage'])).resolve()
        self.save_dfs()

    @classmethod
    def from_model(cls, sim : Simulation, include_crash:bool = False):
        data_dict = {}
        data_dict['params'] = sim.params
        data_dict["n_temporal_units_to_sim"] = sim.n_temporal_units_to_sim
        data_dict["n_temporal_units_simulated"] = sim.n_temporal_units_simulated
        data_dict["temporal_units_by_step"] = sim.params["temporal_units_by_step"]
        data_dict["has_crashed"] = sim.has_crashed
        data_dict["regions"] = sim.model.regions
        data_dict["sectors"] = sim.model.sectors
        with (pathlib.Path(sim.params["results_storage"])/"simulated_events.json").open() as f:
            events = json.load(f)

        data_dict["events"] = events
        data_dict["prod"] = sim.model.production_evolution
        data_dict["prodmax"] = sim.model.production_cap_evolution
        data_dict["overprod"] = sim.model.overproduction_evolution
        data_dict["final_demand"] = sim.model.final_demand_evolution
        data_dict["io_demand"] = sim.model.io_demand_evolution
        data_dict["r_demand"] = sim.model.rebuild_demand_evolution
        data_dict["r_prod"] = sim.model.rebuild_production_evolution
        data_dict["fd_unmet"] = sim.model.final_demand_unmet_evolution
        if sim.params['register_stocks']:
            data_dict["stocks"] = sim.model.stocks_evolution
        data_dict["limiting_stocks"] = sim.model.limiting_stocks_evolution
        return cls(data_dict, include_crash)

    @classmethod
    def from_storage_path(cls, storage_path:str, params=None, include_crash:bool = False) -> Indicators:
        return cls(cls.dict_from_storage_path(storage_path, params=params), include_crash)

    @classmethod
    def from_folder(cls, folder: Union[str, pathlib.Path], indexes_file: Union[str, pathlib.Path], include_crash:bool = False) -> Indicators:
        data_dict = {}
        if not isinstance(indexes_file, pathlib.Path):
            indexes_file = pathlib.Path(indexes_file)
            if not indexes_file.exists():
                raise FileNotFoundError(str("File does not exist:"+str(indexes_file)))
        if not isinstance(folder, pathlib.Path):
            folder = pathlib.Path(folder)
            if not folder.exists():
                raise FileNotFoundError(str("Directory does not exist:"+str(folder)))
        with indexes_file.open() as f:
            indexes = json.load(f)

        params_file = {f.stem : f for f in folder.glob("*.json")}
        absentee = [f for f in cls.params_list if f not in params_file.keys()]
        if absentee != []:
            raise FileNotFoundError("Some of the required parameters files not found (looked for {} in {}".format(cls.params_list,folder))

        record_files = [f for f in folder.glob("*record") if f.is_file()]
        absentee = [f for f in cls.record_files_list if f not in [fn.name for fn in record_files]]
        if absentee != []:
            raise FileNotFoundError("Some of the required records are not there : {}".format(absentee))

        with params_file['simulated_params'].open('r') as f:
            params = json.load(f)

        with params_file['simulated_events'].open('r') as f:
            events = json.load(f)

        if "has_crashed" in params:
            data_dict["has_crashed"] = params["has_crashed"]
        else:
            data_dict["has_crashed"] = False
        results_path = data_dict["results_storage"] = folder.absolute()
        data_dict["n_temporal_units_to_sim"] = params['n_temporal_units_to_sim']
        t = data_dict["n_temporal_units_to_sim"]
        data_dict['params'] = params
        data_dict["n_temporal_units_simulated"] = params['n_temporal_units_simulated']
        data_dict["temporal_units_by_step"] = params["temporal_units_by_step"]
        data_dict["regions"] = indexes["regions"]
        data_dict["n_regions"] = indexes["n_regions"]
        data_dict["sectors"] = indexes["sectors"]
        data_dict["n_sectors"] = indexes["n_sectors"]
        data_dict["events"] = events
        data_dict["prod"] = np.memmap(results_path/"iotable_XVA_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["prodmax"] = np.memmap(results_path/"iotable_X_max_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["overprod"] = np.memmap(results_path/"overprodvector_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["final_demand"] = np.memmap(results_path/"final_demand_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["io_demand"] = np.memmap(results_path/"io_demand_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["r_demand"] = np.memmap(results_path/"rebuild_demand_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["r_prod"] = np.memmap(results_path/"rebuild_prod_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["fd_unmet"] = np.memmap(results_path/"final_demand_unmet_record", mode='r+', dtype='float64',shape=(t,indexes['n_regions']*indexes['n_sectors']))
        data_dict["limiting_stocks"] = np.memmap(results_path/"limiting_stocks_record", mode='r+', dtype='byte',shape=(t*indexes['n_sectors'],indexes['n_industries']))
        if params['register_stocks']:
            if not (results_path/"stocks_record").exists():
                raise FileNotFoundError("Stocks record file was not found {}".format(results_path/"stocks_record"))
            data_dict["stocks"] = np.memmap(results_path/"stocks_record", mode='r+', dtype='float64',shape=(t*indexes['n_sectors'],indexes['n_industries']))
        return cls(data_dict, include_crash)


    @classmethod
    def dict_from_storage_path(cls, storage_path:Union[str,pathlib.Path], params=None) -> dict:
        data_dict = {}
        if not isinstance(storage_path, pathlib.Path):
            storage_path = pathlib.Path(storage_path)
            assert storage_path.exists(), str("Directory does not exist:"+str(storage_path))
        if params is not None:
            simulation_params = params
        else:
            with (storage_path/"simulated_params.json").open() as f:
                simulation_params = json.load(f)
        if (storage_path/simulation_params['results_storage']/"simulated_params.json").exists():
            with (storage_path/simulation_params['results_storage']/"simulated_params.json").open() as f:
                simulation_params = json.load(f)
        with (storage_path/simulation_params['results_storage']/"indexes.json").open() as f:
            indexes = json.load(f)
        with (storage_path/simulation_params['results_storage']/"simulated_events.json").open() as f:
            events = json.load(f)
        t = simulation_params["n_temporal_units_to_sim"]
        if indexes['fd_cat'] is None:
            indexes['fd_cat'] = np.array(["Final demand"])
        results_path = storage_path/pathlib.Path(simulation_params['results_storage'])
        if "has_crashed" in simulation_params:
            data_dict["has_crashed"] = simulation_params["has_crashed"]
        data_dict['params'] = simulation_params
        data_dict["results_storage"] = results_path
        data_dict["temporal_units_by_step"] = simulation_params["temporal_units_by_step"]
        data_dict["n_temporal_units_to_sim"] = t
        data_dict["n_temporal_units_simulated"] = simulation_params['n_temporal_units_simulated']
        data_dict["regions"] = indexes["regions"]
        data_dict["n_regions"] = indexes["n_regions"]
        data_dict["sectors"] = indexes["sectors"]
        data_dict["n_sectors"] = indexes["n_sectors"]
        data_dict["events"] = events
        data_dict["prod"] = np.memmap(results_path/"iotable_XVA_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["prodmax"] = np.memmap(results_path/"iotable_X_max_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["overprod"] = np.memmap(results_path/"overprodvector_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["final_demand"] = np.memmap(results_path/"final_demand_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["io_demand"] = np.memmap(results_path/"io_demand_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["r_demand"] = np.memmap(results_path/"rebuild_demand_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["r_prod"] = np.memmap(results_path/"rebuild_prod_record", mode='r+', dtype='float64',shape=(t,indexes['n_industries']))
        data_dict["fd_unmet"] = np.memmap(results_path/"final_demand_unmet_record", mode='r+', dtype='float64',shape=(t,indexes['n_regions']*indexes['n_sectors']))
        if simulation_params['register_stocks']:
            data_dict["stocks"] = np.memmap(results_path/"stocks_record", mode='r+', dtype='float64',shape=(t*indexes['n_sectors'],indexes['n_industries']))
        data_dict["limiting_stocks"] = np.memmap(results_path/"limiting_stocks_record", mode='r+', dtype='byte',shape=(t*indexes['n_sectors'],indexes['n_industries']))
        return data_dict

    def calc_top_failing_sect(self):
        pass

    def calc_tot_fd_unmet(self):
        self.indicators['tot_fd_unmet'] = self.df_loss['fdloss'].sum()

    def calc_aff_fd_unmet(self):
        #TODO: check this
        self.indicators['aff_fd_unmet'] = self.df_loss[self.df_loss.region.isin(self.aff_regions)]['fdloss'].sum()

    def calc_rebuild_durations(self):
        rebuilding = self.r_demand_df.reset_index().melt(id_vars ="step", var_name=['region','sector'], value_name="rebuild_demand").groupby('step').sum(numeric_only=True).ne(0).rebuild_demand.to_numpy()
        self.indicators['rebuild_durations'] = [ sum( 1 for _ in group ) for key, group in itertools.groupby( rebuilding ) if key ]

    def calc_recovery_duration(self):
        pass

    def calc_general_shortage(self):
        #TODO: replace hard values by soft.
        n_regions = self.df_limiting.columns.levels[0].size
        n_sectors = self.df_limiting.columns.levels[1].size
        # have only steps in index (next step not possible with multiindex)
        a = self.df_limiting.unstack()
        # select only simulated steps (we can't store nan in bool or byte dtype array)
        a = a.iloc[lambda x : x.index % self.temporal_units_by_step == 0]
        # We put -1 initially, so this should check we have correctly selected the simulated rows
        assert (a >= 0).all().all()
        # put input stocks as columns and the rest as index
        a = a.stack().T.stack(level=0)
        # sum for all input and divide by n_sector to get "input shortage fraction"
        # by industry
        b = a.sum(axis=1).groupby(['step','region','sector']).sum()/n_sectors
        # by sector
        c = b.groupby(['step','region']).sum()/n_sectors
        # by region
        c = c.groupby('step').sum()/n_regions
        if not c.ne(0).any():
            self.indicators['shortage_b'] = False
        else:
            self.indicators['shortage_b'] = True
            shortage_date_start = c.ne(0.0).idxmax()
            self.indicators['shortage_date_start'] = shortage_date_start
            shortage_date_end = c.loc[shortage_date_start:].eq(0).idxmax()
            self.indicators['shortage_date_end'] = shortage_date_end
            self.indicators['shortage_date_max'] = c.idxmax()
            self.indicators['shortage_ind_max'] = c.max()
            self.indicators['shortage_ind_mean'] = c.loc[shortage_date_start:shortage_date_end].mean()

    def calc_first_shortages(self):
        a = self.df_limiting.stack([0,1]) #type: ignore
        a = a.swaplevel(1,2).swaplevel(2,3)
        b = a[a>0]
        b = b[:10]
        res = list(b.index) #type:ignore
        self.indicators['10_first_shortages_(step,region,sector,stock_of)'] = res

    def calc_tot_prod_change(self):
        df2=self.prod_df.copy()
        #df2.columns=df2.columns.droplevel(0)
        prod_chg = df2 - df2.iloc[0,:]
        # Round to € to avoid floating error differences
        prod_chg = prod_chg.round(6)
        # Aggregate rebuilding and non-rebuilding sectors
        prod_chg_agg_1 = pd.concat([prod_chg.loc[:,pd.IndexSlice[:,self.rebuilding_sectors]].groupby('region',axis=1).sum(),
                                    prod_chg.loc[prod_chg.index.difference(prod_chg.loc[:,pd.IndexSlice[:,self.rebuilding_sectors]].columns)].groupby('region',axis=1).sum()],
                                   keys=['rebuilding', 'non-rebuilding'], names=['sectors affected'],axis=1)

        n_semesters = (self.n_rows // (self.params['year_to_temporal_unit_factor'] // 2))
        row_to_semester = (self.params['year_to_temporal_unit_factor'] // 2)
        modulo = (self.params['year_to_temporal_unit_factor'] % 2)
        logger.info("There are {} semesters [{} rows, each representing a {}th of a year]".format(n_semesters, self.n_rows, self.params['year_to_temporal_unit_factor']))
        prod_chg_sect_sem_l = []
        for sem in range(0,n_semesters):
            prod_chg_sect_sem_l.append(prod_chg_agg_1.iloc[sem*row_to_semester:((sem+1)*row_to_semester)+(sem%2)*modulo].sum().T)

        prod_chg_region = pd.concat(prod_chg_sect_sem_l, keys=['semester {}'.format(v+1) for v in range(n_semesters)], names=['semester'])
        prod_chg_region = pd.DataFrame({(self.storage_path.parent.name,self.storage_path.name):prod_chg_region}).T
        prod_chg_region.to_json(self.storage_path/"prod_chg.json", indent=4, orient='split')

        prod_chg_sect = prod_chg.sum()
        tmp = prod_chg_sect.sort_values(ascending=False,key=abs).head(5).to_dict()
        self.indicators['top_5_sector_chg'] = {str(k): v for k, v in tmp.items()}

        self.indicators['prod_gain_tot'] = prod_chg.mul(prod_chg.gt(0)).sum().sum()
        self.indicators['prod_lost_tot'] = prod_chg.mul(~prod_chg.gt(0)).sum().sum() * (-1)
        prod_chg = prod_chg.drop(self.aff_regions, axis=1)
        self.indicators['prod_gain_unaff'] = prod_chg.mul(prod_chg.gt(0)).sum().sum()
        self.indicators['prod_lost_unaff'] = prod_chg.mul(~prod_chg.gt(0)).sum().sum() * (-1)

    def update_indicators(self):
        logger.info("(Re)computing all indicators")
        logger.info("Tot fd unmet")
        self.calc_tot_fd_unmet()
        logger.info("aff fd unmet")
        self.calc_aff_fd_unmet()
        logger.info("rebuild durations")
        self.calc_rebuild_durations()
        logger.info("recovery duration")
        self.calc_recovery_duration()
        logger.info("general shortage")
        self.calc_general_shortage()
        logger.info("tot prod change")
        self.calc_tot_prod_change()
        logger.info("fd loss region")
        self.calc_fd_loss_region()
        logger.info("first shortages")
        self.calc_first_shortages()

    def write_indicators(self):
        logger.info("Writing indicators to json")
        #self.update_indicators()
        with self.storage.open('w') as f:
            json.dump(self.indicators, f, cls=numpyencoder.NumpyEncoder)

    def calc_fd_loss_region(self):
        fd_loss = self.fd_unmet_df.copy().round(6)
        fd_loss_agg_1 = pd.concat([fd_loss.loc[:,pd.IndexSlice[:,self.rebuilding_sectors]].groupby('region',axis=1).sum(),
                                    fd_loss.loc[fd_loss.index.difference(fd_loss.loc[:,pd.IndexSlice[:,self.rebuilding_sectors]].columns)].groupby('region',axis=1).sum()],
                                   keys=['rebuilding', 'non-rebuilding'], names=['sectors affected'],axis=1)

        n_semesters = (self.n_rows // (self.params['year_to_temporal_unit_factor'] // 2))
        row_to_semester = (self.params['year_to_temporal_unit_factor'] // 2)
        modulo = (self.params['year_to_temporal_unit_factor'] % 2)
        logger.info("There are {} semesters [{} rows, each representing a {}th of a year]".format(n_semesters, self.n_rows, self.params['year_to_temporal_unit_factor']))
        fd_loss_sect_sem_l = []
        for sem in range(0,n_semesters):
            fd_loss_sect_sem_l.append(fd_loss_agg_1.iloc[sem*row_to_semester:((sem+1)*row_to_semester)+(sem%2)*modulo].sum().T)

        fd_loss_region = pd.concat(fd_loss_sect_sem_l, keys=['semester {}'.format(v+1) for v in range(n_semesters)], names=['semester'])
        fd_loss_region = pd.DataFrame({(self.storage_path.parent.name,self.storage_path.name):fd_loss_region}).T
        fd_loss_region.to_json(self.storage_path/"fd_loss.json", indent=4, orient='split')

        fd_loss_sect = fd_loss.sum()
        tmp = fd_loss_sect.sort_values(ascending=False,key=abs).head(5).to_dict()
        self.indicators['top_5_sector_fdloss'] = {str(k): v for k, v in tmp.items()}

    def save_dfs(self):
        logger.info("Saving computed dataframe to results folder")
        self.prod_df.to_parquet(self.storage_path/"prod_df.parquet")
        self.prodmax_df.to_parquet(self.storage_path/"prodmax_df.parquet")
        self.overprod_df.to_parquet(self.storage_path/"overprod_df.parquet")
        self.fd_unmet_df.to_parquet(self.storage_path/"fd_unmet_df.parquet")
        self.final_demand_df.to_parquet(self.storage_path/"final_demand_df.parquet")
        self.io_demand_df.to_parquet(self.storage_path/"io_demand_df.parquet")
        self.r_demand_df.to_parquet(self.storage_path/"r_demand_df.parquet")
        self.df_loss.to_parquet(self.storage_path/"treated_df_loss.parquet")
        if self.df_stocks is not None:
            ddf = da.from_pandas(self.df_stocks, chunksize=10000000)
            ddf.to_parquet(self.storage_path/"treated_df_stocks.parquet", engine="pyarrow")
        if self.df_limiting is not None:
            #ddf_l = da.from_pandas(self.df_limiting, chunksize=10000000)
            #ddf_l = ddf_l.melt(ignore_index=False).rename(columns={'variable_0':'region','variable_1':'sector', 'variable_2':'stock of'})
            #ddf_l = ddf_l.reset_index()
            #ddf_l['step'] = ddf_l['step'].astype("uint16")
            #ddf_l['stock of'] = ddf_l['stock of'].astype("category")
            #ddf_l['region'] = ddf_l['region'].astype("category")
            #ddf_l['sector'] = ddf_l['sector'].astype("category")
            self.df_limiting.to_parquet(self.storage_path/"treated_df_limiting.parquet", engine="pyarrow")
        #self.df_limiting.to_feather(self.storage_path/"treated_df_limiting.feather")



"""
This type stub file was generated by pyright.
"""

from typing import TYPE_CHECKING
from pandas._typing import ReadBuffer
from pandas.io.parsers.base_parser import ParserBase
from pandas import DataFrame

if TYPE_CHECKING:
    ...
class ArrowParserWrapper(ParserBase):
    """
    Wrapper for the pyarrow engine for read_csv()
    """
    def __init__(self, src: ReadBuffer[bytes], **kwds) -> None:
        ...
    
    def read(self) -> DataFrame:
        """
        Reads the contents of a CSV file into a DataFrame and
        processes it according to the kwargs passed in the
        constructor.

        Returns
        -------
        DataFrame
            The DataFrame created from the CSV file.
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

from abc import ABCMeta, abstractmethod
from typing import Hashable, TYPE_CHECKING, overload
from pandas._typing import CompressionOptions, FilePath, ReadBuffer
from pandas.util._decorators import deprecate_nonkeyword_arguments, doc
from pandas.core.shared_docs import _shared_docs
from pandas import DataFrame

"""
Read SAS sas7bdat or xport files.
"""
if TYPE_CHECKING:
    ...
class ReaderBase(metaclass=ABCMeta):
    """
    Protocol for XportReader and SAS7BDATReader classes.
    """
    @abstractmethod
    def read(self, nrows: int | None = ...) -> DataFrame:
        ...
    
    @abstractmethod
    def close(self) -> None:
        ...
    
    def __enter__(self) -> ReaderBase:
        ...
    
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...
    


@overload
def read_sas(filepath_or_buffer: FilePath | ReadBuffer[bytes], format: str | None = ..., index: Hashable | None = ..., encoding: str | None = ..., chunksize: int = ..., iterator: bool = ..., compression: CompressionOptions = ...) -> ReaderBase:
    ...

@overload
def read_sas(filepath_or_buffer: FilePath | ReadBuffer[bytes], format: str | None = ..., index: Hashable | None = ..., encoding: str | None = ..., chunksize: None = ..., iterator: bool = ..., compression: CompressionOptions = ...) -> DataFrame | ReaderBase:
    ...

@deprecate_nonkeyword_arguments(version=None, allowed_args=["filepath_or_buffer"])
@doc(decompression_options=_shared_docs["decompression_options"] % "filepath_or_buffer")
def read_sas(filepath_or_buffer: FilePath | ReadBuffer[bytes], format: str | None = ..., index: Hashable | None = ..., encoding: str | None = ..., chunksize: int | None = ..., iterator: bool = ..., compression: CompressionOptions = ...) -> DataFrame | ReaderBase:
    """
    Read SAS files stored as either XPORT or SAS7BDAT format files.

    Parameters
    ----------
    filepath_or_buffer : str, path object, or file-like object
        String, path object (implementing ``os.PathLike[str]``), or file-like
        object implementing a binary ``read()`` function. The string could be a URL.
        Valid URL schemes include http, ftp, s3, and file. For file URLs, a host is
        expected. A local file could be:
        ``file://localhost/path/to/table.sas``.
    format : str {{'xport', 'sas7bdat'}} or None
        If None, file format is inferred from file extension. If 'xport' or
        'sas7bdat', uses the corresponding format.
    index : identifier of index column, defaults to None
        Identifier of column that should be used as index of the DataFrame.
    encoding : str, default is None
        Encoding for text data.  If None, text data are stored as raw bytes.
    chunksize : int
        Read file `chunksize` lines at a time, returns iterator.

        .. versionchanged:: 1.2

            ``TextFileReader`` is a context manager.
    iterator : bool, defaults to False
        If True, returns an iterator for reading the file incrementally.

        .. versionchanged:: 1.2

            ``TextFileReader`` is a context manager.
    {decompression_options}

    Returns
    -------
    DataFrame if iterator=False and chunksize=None, else SAS7BDATReader
    or XportReader
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any, Callable, Final, Generator, Hashable, Literal, overload
from pandas._typing import ArrayLike, DtypeObj, npt
from enum import Enum

ndarray_obj_2d = np.ndarray
class _NoDefault(Enum):
    no_default = ...


no_default: Final = ...
NoDefault = Literal[_NoDefault.no_default]
i8max: int
u8max: int
def item_from_zerodim(val: object) -> object:
    ...

def infer_dtype(value: object, skipna: bool = ...) -> str:
    ...

def is_iterator(obj: object) -> bool:
    ...

def is_scalar(val: object) -> bool:
    ...

def is_list_like(obj: object, allow_sets: bool = ...) -> bool:
    ...

def is_period(val: object) -> bool:
    ...

def is_interval(val: object) -> bool:
    ...

def is_decimal(val: object) -> bool:
    ...

def is_complex(val: object) -> bool:
    ...

def is_bool(val: object) -> bool:
    ...

def is_integer(val: object) -> bool:
    ...

def is_float(val: object) -> bool:
    ...

def is_interval_array(values: np.ndarray) -> bool:
    ...

def is_datetime64_array(values: np.ndarray) -> bool:
    ...

def is_timedelta_or_timedelta64_array(values: np.ndarray) -> bool:
    ...

def is_datetime_with_singletz_array(values: np.ndarray) -> bool:
    ...

def is_time_array(values: np.ndarray, skipna: bool = ...):
    ...

def is_date_array(values: np.ndarray, skipna: bool = ...):
    ...

def is_datetime_array(values: np.ndarray, skipna: bool = ...):
    ...

def is_string_array(values: np.ndarray, skipna: bool = ...):
    ...

def is_float_array(values: np.ndarray, skipna: bool = ...):
    ...

def is_integer_array(values: np.ndarray, skipna: bool = ...):
    ...

def is_bool_array(values: np.ndarray, skipna: bool = ...):
    ...

def fast_multiget(mapping: dict, keys: np.ndarray, default=...) -> np.ndarray:
    ...

def fast_unique_multiple_list_gen(gen: Generator, sort: bool = ...) -> list:
    ...

def fast_unique_multiple_list(lists: list, sort: bool | None = ...) -> list:
    ...

def fast_unique_multiple(arrays: list, sort: bool = ...) -> list:
    ...

def map_infer(arr: np.ndarray, f: Callable[[Any], Any], convert: bool = ..., ignore_na: bool = ...) -> np.ndarray:
    ...

@overload
def maybe_convert_objects(objects: npt.NDArray[np.object_], *, try_float: bool = ..., safe: bool = ..., convert_datetime: Literal[False] = ..., convert_timedelta: bool = ..., convert_period: Literal[False] = ..., convert_interval: Literal[False] = ..., convert_to_nullable_integer: Literal[False] = ..., dtype_if_all_nat: DtypeObj | None = ...) -> np.ndarray:
    ...

@overload
def maybe_convert_objects(objects: npt.NDArray[np.object_], *, try_float: bool = ..., safe: bool = ..., convert_datetime: bool = ..., convert_timedelta: bool = ..., convert_period: bool = ..., convert_interval: bool = ..., convert_to_nullable_integer: Literal[True] = ..., dtype_if_all_nat: DtypeObj | None = ...) -> ArrayLike:
    ...

@overload
def maybe_convert_objects(objects: npt.NDArray[np.object_], *, try_float: bool = ..., safe: bool = ..., convert_datetime: Literal[True] = ..., convert_timedelta: bool = ..., convert_period: bool = ..., convert_interval: bool = ..., convert_to_nullable_integer: bool = ..., dtype_if_all_nat: DtypeObj | None = ...) -> ArrayLike:
    ...

@overload
def maybe_convert_objects(objects: npt.NDArray[np.object_], *, try_float: bool = ..., safe: bool = ..., convert_datetime: bool = ..., convert_timedelta: bool = ..., convert_period: Literal[True] = ..., convert_interval: bool = ..., convert_to_nullable_integer: bool = ..., dtype_if_all_nat: DtypeObj | None = ...) -> ArrayLike:
    ...

@overload
def maybe_convert_objects(objects: npt.NDArray[np.object_], *, try_float: bool = ..., safe: bool = ..., convert_datetime: bool = ..., convert_timedelta: bool = ..., convert_period: bool = ..., convert_interval: bool = ..., convert_to_nullable_integer: bool = ..., dtype_if_all_nat: DtypeObj | None = ...) -> ArrayLike:
    ...

@overload
def maybe_convert_numeric(values: npt.NDArray[np.object_], na_values: set, convert_empty: bool = ..., coerce_numeric: bool = ..., convert_to_masked_nullable: Literal[False] = ...) -> tuple[np.ndarray, None]:
    ...

@overload
def maybe_convert_numeric(values: npt.NDArray[np.object_], na_values: set, convert_empty: bool = ..., coerce_numeric: bool = ..., *, convert_to_masked_nullable: Literal[True]) -> tuple[np.ndarray, np.ndarray]:
    ...

def ensure_string_array(arr, na_value: object = ..., convert_na_value: bool = ..., copy: bool = ..., skipna: bool = ...) -> npt.NDArray[np.object_]:
    ...

def infer_datetimelike_array(arr: npt.NDArray[np.object_]) -> tuple[str, bool]:
    ...

def convert_nans_to_NA(arr: npt.NDArray[np.object_]) -> npt.NDArray[np.object_]:
    ...

def fast_zip(ndarrays: list) -> npt.NDArray[np.object_]:
    ...

def to_object_array_tuples(rows: object) -> ndarray_obj_2d:
    ...

def tuples_to_object_array(tuples: npt.NDArray[np.object_]) -> ndarray_obj_2d:
    ...

def to_object_array(rows: object, min_width: int = ...) -> ndarray_obj_2d:
    ...

def dicts_to_array(dicts: list, columns: list) -> ndarray_obj_2d:
    ...

def maybe_booleans_to_slice(mask: npt.NDArray[np.uint8]) -> slice | npt.NDArray[np.uint8]:
    ...

def maybe_indices_to_slice(indices: npt.NDArray[np.intp], max_len: int) -> slice | npt.NDArray[np.intp]:
    ...

def is_all_arraylike(obj: list) -> bool:
    ...

def memory_usage_of_objects(arr: np.ndarray) -> int:
    ...

def map_infer_mask(arr: np.ndarray, f: Callable[[Any], Any], mask: np.ndarray, convert: bool = ..., na_value: Any = ..., dtype: np.dtype = ...) -> np.ndarray:
    ...

def indices_fast(index: npt.NDArray[np.intp], labels: np.ndarray, keys: list, sorted_labels: list[npt.NDArray[np.int64]]) -> dict[Hashable, npt.NDArray[np.intp]]:
    ...

def generate_slices(labels: np.ndarray, ngroups: int) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.int64]]:
    ...

def count_level_2d(mask: np.ndarray, labels: np.ndarray, max_bin: int, axis: int) -> np.ndarray:
    ...

def get_level_sorter(label: np.ndarray, starts: np.ndarray) -> np.ndarray:
    ...

def generate_bins_dt64(values: npt.NDArray[np.int64], binner: np.ndarray, closed: object = ..., hasnans: bool = ...) -> np.ndarray:
    ...

def array_equivalent_object(left: np.ndarray, right: np.ndarray) -> bool:
    ...

def has_infs(arr: np.ndarray) -> bool:
    ...

def get_reverse_indexer(indexer: np.ndarray, length: int) -> npt.NDArray[np.intp]:
    ...

def is_bool_list(obj: list) -> bool:
    ...

def dtypes_all_equal(types: list[DtypeObj]) -> bool:
    ...




"""
This type stub file was generated by pyright.
"""

from abc import ABC, abstractmethod
from collections import abc
from typing import Any, Callable, Generic, Literal, Mapping, TYPE_CHECKING, TypeVar, overload
from pandas._typing import CompressionOptions, DtypeArg, FilePath, IndexLabel, JSONSerializable, ReadBuffer, StorageOptions, WriteBuffer
from pandas.util._decorators import deprecate_kwarg, deprecate_nonkeyword_arguments, doc
from pandas import DataFrame, Series
from pandas.core.shared_docs import _shared_docs
from pandas.core.generic import NDFrame

if TYPE_CHECKING:
    ...
FrameSeriesStrT = TypeVar("FrameSeriesStrT", bound=Literal["frame", "series"])
loads = ...
dumps = ...
@overload
def to_json(path_or_buf: FilePath | WriteBuffer[str] | WriteBuffer[bytes], obj: NDFrame, orient: str | None = ..., date_format: str = ..., double_precision: int = ..., force_ascii: bool = ..., date_unit: str = ..., default_handler: Callable[[Any], JSONSerializable] | None = ..., lines: bool = ..., compression: CompressionOptions = ..., index: bool = ..., indent: int = ..., storage_options: StorageOptions = ...) -> None:
    ...

@overload
def to_json(path_or_buf: None, obj: NDFrame, orient: str | None = ..., date_format: str = ..., double_precision: int = ..., force_ascii: bool = ..., date_unit: str = ..., default_handler: Callable[[Any], JSONSerializable] | None = ..., lines: bool = ..., compression: CompressionOptions = ..., index: bool = ..., indent: int = ..., storage_options: StorageOptions = ...) -> str:
    ...

def to_json(path_or_buf: FilePath | WriteBuffer[str] | WriteBuffer[bytes] | None, obj: NDFrame, orient: str | None = ..., date_format: str = ..., double_precision: int = ..., force_ascii: bool = ..., date_unit: str = ..., default_handler: Callable[[Any], JSONSerializable] | None = ..., lines: bool = ..., compression: CompressionOptions = ..., index: bool = ..., indent: int = ..., storage_options: StorageOptions = ...) -> str | None:
    ...

class Writer(ABC):
    _default_orient: str
    def __init__(self, obj, orient: str | None, date_format: str, double_precision: int, ensure_ascii: bool, date_unit: str, index: bool, default_handler: Callable[[Any], JSONSerializable] | None = ..., indent: int = ...) -> None:
        ...
    
    def write(self) -> str:
        ...
    
    @property
    @abstractmethod
    def obj_to_write(self) -> NDFrame | Mapping[IndexLabel, Any]:
        """Object to write in JSON format."""
        ...
    


class SeriesWriter(Writer):
    _default_orient = ...
    @property
    def obj_to_write(self) -> NDFrame | Mapping[IndexLabel, Any]:
        ...
    


class FrameWriter(Writer):
    _default_orient = ...
    @property
    def obj_to_write(self) -> NDFrame | Mapping[IndexLabel, Any]:
        ...
    


class JSONTableWriter(FrameWriter):
    _default_orient = ...
    def __init__(self, obj, orient: str | None, date_format: str, double_precision: int, ensure_ascii: bool, date_unit: str, index: bool, default_handler: Callable[[Any], JSONSerializable] | None = ..., indent: int = ...) -> None:
        """
        Adds a `schema` attribute with the Table Schema, resets
        the index (can't do in caller, because the schema inference needs
        to know what the index is, forces orient to records, and forces
        date_format to 'iso'.
        """
        ...
    
    @property
    def obj_to_write(self) -> NDFrame | Mapping[IndexLabel, Any]:
        ...
    


@overload
def read_json(path_or_buf: FilePath | ReadBuffer[str] | ReadBuffer[bytes], *, orient: str | None = ..., typ: Literal["frame"] = ..., dtype: DtypeArg | None = ..., convert_axes=..., convert_dates: bool | list[str] = ..., keep_default_dates: bool = ..., numpy: bool = ..., precise_float: bool = ..., date_unit: str | None = ..., encoding: str | None = ..., encoding_errors: str | None = ..., lines: bool = ..., chunksize: int, compression: CompressionOptions = ..., nrows: int | None = ..., storage_options: StorageOptions = ...) -> JsonReader[Literal["frame"]]:
    ...

@overload
def read_json(path_or_buf: FilePath | ReadBuffer[str] | ReadBuffer[bytes], *, orient: str | None = ..., typ: Literal["series"], dtype: DtypeArg | None = ..., convert_axes=..., convert_dates: bool | list[str] = ..., keep_default_dates: bool = ..., numpy: bool = ..., precise_float: bool = ..., date_unit: str | None = ..., encoding: str | None = ..., encoding_errors: str | None = ..., lines: bool = ..., chunksize: int, compression: CompressionOptions = ..., nrows: int | None = ..., storage_options: StorageOptions = ...) -> JsonReader[Literal["series"]]:
    ...

@overload
def read_json(path_or_buf: FilePath | ReadBuffer[str] | ReadBuffer[bytes], *, orient: str | None = ..., typ: Literal["series"], dtype: DtypeArg | None = ..., convert_axes=..., convert_dates: bool | list[str] = ..., keep_default_dates: bool = ..., numpy: bool = ..., precise_float: bool = ..., date_unit: str | None = ..., encoding: str | None = ..., encoding_errors: str | None = ..., lines: bool = ..., chunksize: None = ..., compression: CompressionOptions = ..., nrows: int | None = ..., storage_options: StorageOptions = ...) -> Series:
    ...

@overload
def read_json(path_or_buf: FilePath | ReadBuffer[str] | ReadBuffer[bytes], orient: str | None = ..., typ: Literal["frame"] = ..., dtype: DtypeArg | None = ..., convert_axes=..., convert_dates: bool | list[str] = ..., keep_default_dates: bool = ..., numpy: bool = ..., precise_float: bool = ..., date_unit: str | None = ..., encoding: str | None = ..., encoding_errors: str | None = ..., lines: bool = ..., chunksize: None = ..., compression: CompressionOptions = ..., nrows: int | None = ..., storage_options: StorageOptions = ...) -> DataFrame:
    ...

@doc(storage_options=_shared_docs["storage_options"], decompression_options=_shared_docs["decompression_options"] % "path_or_buf")
@deprecate_kwarg(old_arg_name="numpy", new_arg_name=None)
@deprecate_nonkeyword_arguments(version="2.0", allowed_args=["path_or_buf"])
def read_json(path_or_buf: FilePath | ReadBuffer[str] | ReadBuffer[bytes], orient: str | None = ..., typ: Literal["frame", "series"] = ..., dtype: DtypeArg | None = ..., convert_axes=..., convert_dates: bool | list[str] = ..., keep_default_dates: bool = ..., numpy: bool = ..., precise_float: bool = ..., date_unit: str | None = ..., encoding: str | None = ..., encoding_errors: str | None = ..., lines: bool = ..., chunksize: int | None = ..., compression: CompressionOptions = ..., nrows: int | None = ..., storage_options: StorageOptions = ...) -> DataFrame | Series | JsonReader:
    """
    Convert a JSON string to pandas object.

    Parameters
    ----------
    path_or_buf : a valid JSON str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL. Valid
        URL schemes include http, ftp, s3, and file. For file URLs, a host is
        expected. A local file could be:
        ``file://localhost/path/to/table.json``.

        If you want to pass in a path object, pandas accepts any
        ``os.PathLike``.

        By file-like object, we refer to objects with a ``read()`` method,
        such as a file handle (e.g. via builtin ``open`` function)
        or ``StringIO``.
    orient : str
        Indication of expected JSON string format.
        Compatible JSON strings can be produced by ``to_json()`` with a
        corresponding orient value.
        The set of possible orients is:

        - ``'split'`` : dict like
          ``{{index -> [index], columns -> [columns], data -> [values]}}``
        - ``'records'`` : list like
          ``[{{column -> value}}, ... , {{column -> value}}]``
        - ``'index'`` : dict like ``{{index -> {{column -> value}}}}``
        - ``'columns'`` : dict like ``{{column -> {{index -> value}}}}``
        - ``'values'`` : just the values array

        The allowed and default values depend on the value
        of the `typ` parameter.

        * when ``typ == 'series'``,

          - allowed orients are ``{{'split','records','index'}}``
          - default is ``'index'``
          - The Series index must be unique for orient ``'index'``.

        * when ``typ == 'frame'``,

          - allowed orients are ``{{'split','records','index',
            'columns','values', 'table'}}``
          - default is ``'columns'``
          - The DataFrame index must be unique for orients ``'index'`` and
            ``'columns'``.
          - The DataFrame columns must be unique for orients ``'index'``,
            ``'columns'``, and ``'records'``.

    typ : {{'frame', 'series'}}, default 'frame'
        The type of object to recover.

    dtype : bool or dict, default None
        If True, infer dtypes; if a dict of column to dtype, then use those;
        if False, then don't infer dtypes at all, applies only to the data.

        For all ``orient`` values except ``'table'``, default is True.

        .. versionchanged:: 0.25.0

           Not applicable for ``orient='table'``.

    convert_axes : bool, default None
        Try to convert the axes to the proper dtypes.

        For all ``orient`` values except ``'table'``, default is True.

        .. versionchanged:: 0.25.0

           Not applicable for ``orient='table'``.

    convert_dates : bool or list of str, default True
        If True then default datelike columns may be converted (depending on
        keep_default_dates).
        If False, no dates will be converted.
        If a list of column names, then those columns will be converted and
        default datelike columns may also be converted (depending on
        keep_default_dates).

    keep_default_dates : bool, default True
        If parsing dates (convert_dates is not False), then try to parse the
        default datelike columns.
        A column label is datelike if

        * it ends with ``'_at'``,

        * it ends with ``'_time'``,

        * it begins with ``'timestamp'``,

        * it is ``'modified'``, or

        * it is ``'date'``.

    numpy : bool, default False
        Direct decoding to numpy arrays. Supports numeric data only, but
        non-numeric column and index labels are supported. Note also that the
        JSON ordering MUST be the same for each term if numpy=True.

        .. deprecated:: 1.0.0

    precise_float : bool, default False
        Set to enable usage of higher precision (strtod) function when
        decoding string to double values. Default (False) is to use fast but
        less precise builtin functionality.

    date_unit : str, default None
        The timestamp unit to detect if converting dates. The default behaviour
        is to try and detect the correct precision, but if this is not desired
        then pass one of 's', 'ms', 'us' or 'ns' to force parsing only seconds,
        milliseconds, microseconds or nanoseconds respectively.

    encoding : str, default is 'utf-8'
        The encoding to use to decode py3 bytes.

    encoding_errors : str, optional, default "strict"
        How encoding errors are treated. `List of possible values
        <https://docs.python.org/3/library/codecs.html#error-handlers>`_ .

        .. versionadded:: 1.3.0

    lines : bool, default False
        Read the file as a json object per line.

    chunksize : int, optional
        Return JsonReader object for iteration.
        See the `line-delimited json docs
        <https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#line-delimited-json>`_
        for more information on ``chunksize``.
        This can only be passed if `lines=True`.
        If this is None, the file will be read into memory all at once.

        .. versionchanged:: 1.2

           ``JsonReader`` is a context manager.

    {decompression_options}

        .. versionchanged:: 1.4.0 Zstandard support.

    nrows : int, optional
        The number of lines from the line-delimited jsonfile that has to be read.
        This can only be passed if `lines=True`.
        If this is None, all the rows will be returned.

        .. versionadded:: 1.1

    {storage_options}

        .. versionadded:: 1.2.0

    Returns
    -------
    Series or DataFrame
        The type returned depends on the value of `typ`.

    See Also
    --------
    DataFrame.to_json : Convert a DataFrame to a JSON string.
    Series.to_json : Convert a Series to a JSON string.
    json_normalize : Normalize semi-structured JSON data into a flat table.

    Notes
    -----
    Specific to ``orient='table'``, if a :class:`DataFrame` with a literal
    :class:`Index` name of `index` gets written with :func:`to_json`, the
    subsequent read operation will incorrectly set the :class:`Index` name to
    ``None``. This is because `index` is also used by :func:`DataFrame.to_json`
    to denote a missing :class:`Index` name, and the subsequent
    :func:`read_json` operation cannot distinguish between the two. The same
    limitation is encountered with a :class:`MultiIndex` and any names
    beginning with ``'level_'``.

    Examples
    --------
    >>> df = pd.DataFrame([['a', 'b'], ['c', 'd']],
    ...                   index=['row 1', 'row 2'],
    ...                   columns=['col 1', 'col 2'])

    Encoding/decoding a Dataframe using ``'split'`` formatted JSON:

    >>> df.to_json(orient='split')
        '\
{{\
"columns":["col 1","col 2"],\
"index":["row 1","row 2"],\
"data":[["a","b"],["c","d"]]\
}}\
'
    >>> pd.read_json(_, orient='split')
          col 1 col 2
    row 1     a     b
    row 2     c     d

    Encoding/decoding a Dataframe using ``'index'`` formatted JSON:

    >>> df.to_json(orient='index')
    '{{"row 1":{{"col 1":"a","col 2":"b"}},"row 2":{{"col 1":"c","col 2":"d"}}}}'

    >>> pd.read_json(_, orient='index')
          col 1 col 2
    row 1     a     b
    row 2     c     d

    Encoding/decoding a Dataframe using ``'records'`` formatted JSON.
    Note that index labels are not preserved with this encoding.

    >>> df.to_json(orient='records')
    '[{{"col 1":"a","col 2":"b"}},{{"col 1":"c","col 2":"d"}}]'
    >>> pd.read_json(_, orient='records')
      col 1 col 2
    0     a     b
    1     c     d

    Encoding with Table Schema

    >>> df.to_json(orient='table')
        '\
{{"schema":{{"fields":[\
{{"name":"index","type":"string"}},\
{{"name":"col 1","type":"string"}},\
{{"name":"col 2","type":"string"}}],\
"primaryKey":["index"],\
"pandas_version":"1.4.0"}},\
"data":[\
{{"index":"row 1","col 1":"a","col 2":"b"}},\
{{"index":"row 2","col 1":"c","col 2":"d"}}]\
}}\
'
    """
    ...

class JsonReader(abc.Iterator, Generic[FrameSeriesStrT]):
    """
    JsonReader provides an interface for reading in a JSON file.

    If initialized with ``lines=True`` and ``chunksize``, can be iterated over
    ``chunksize`` lines at a time. Otherwise, calling ``read`` reads in the
    whole document.
    """
    def __init__(self, filepath_or_buffer, orient, typ: FrameSeriesStrT, dtype, convert_axes, convert_dates, keep_default_dates: bool, numpy: bool, precise_float: bool, date_unit, encoding, lines: bool, chunksize: int | None, compression: CompressionOptions, nrows: int | None, storage_options: StorageOptions = ..., encoding_errors: str | None = ...) -> None:
        ...
    
    @overload
    def read(self: JsonReader[Literal["frame"]]) -> DataFrame:
        ...
    
    @overload
    def read(self: JsonReader[Literal["series"]]) -> Series:
        ...
    
    @overload
    def read(self: JsonReader[Literal["frame", "series"]]) -> DataFrame | Series:
        ...
    
    def read(self) -> DataFrame | Series:
        """
        Read the whole JSON input into a pandas object.
        """
        ...
    
    def close(self) -> None:
        """
        If we opened a stream earlier, in _get_data_from_filepath, we should
        close it.

        If an open stream or file was passed, we leave it open.
        """
        ...
    
    def __iter__(self: JsonReader[FrameSeriesStrT]) -> JsonReader[FrameSeriesStrT]:
        ...
    
    @overload
    def __next__(self: JsonReader[Literal["frame"]]) -> DataFrame:
        ...
    
    @overload
    def __next__(self: JsonReader[Literal["series"]]) -> Series:
        ...
    
    @overload
    def __next__(self: JsonReader[Literal["frame", "series"]]) -> DataFrame | Series:
        ...
    
    def __next__(self) -> DataFrame | Series:
        ...
    
    def __enter__(self) -> JsonReader[FrameSeriesStrT]:
        ...
    
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...
    


class Parser:
    _split_keys: tuple[str, ...]
    _default_orient: str
    _STAMP_UNITS = ...
    _MIN_STAMPS = ...
    def __init__(self, json, orient, dtype: DtypeArg | None = ..., convert_axes: bool = ..., convert_dates: bool | list[str] = ..., keep_default_dates: bool = ..., numpy: bool = ..., precise_float: bool = ..., date_unit=...) -> None:
        ...
    
    def check_keys_split(self, decoded) -> None:
        """
        Checks that dict has only the appropriate keys for orient='split'.
        """
        ...
    
    def parse(self):
        ...
    


class SeriesParser(Parser):
    _default_orient = ...
    _split_keys = ...


class FrameParser(Parser):
    _default_orient = ...
    _split_keys = ...





"""
This type stub file was generated by pyright.
"""

import contextlib
import functools
from typing import Iterator

@contextlib.contextmanager
def rewrite_exception(old_name: str, new_name: str) -> Iterator[None]:
    """
    Rewrite the message of an exception.
    """
    ...

@functools.lru_cache
def find_stack_level(frame) -> int:
    """
    Find the first place in the stack that is not inside pandas
    (tests notwithstanding).

    ``frame`` should be passed as ``inspect.currentframe()`` by the
    calling function.

    https://stackoverflow.com/questions/17407119/python-inspect-stack-is-slow
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import TYPE_CHECKING
from pandas.core.accessor import PandasDelegate, delegate_names
from pandas.core.arrays import DatetimeArray, PeriodArray, TimedeltaArray
from pandas.core.base import NoNewAttributesMixin, PandasObject
from pandas import DataFrame, Series

"""
datetimelike delegation
"""
if TYPE_CHECKING:
    ...
class Properties(PandasDelegate, PandasObject, NoNewAttributesMixin):
    _hidden_attrs = ...
    def __init__(self, data: Series, orig) -> None:
        ...
    


@delegate_names(delegate=DatetimeArray, accessors=DatetimeArray._datetimelike_ops, typ="property")
@delegate_names(delegate=DatetimeArray, accessors=DatetimeArray._datetimelike_methods, typ="method")
class DatetimeProperties(Properties):
    """
    Accessor object for datetimelike properties of the Series values.

    Examples
    --------
    >>> seconds_series = pd.Series(pd.date_range("2000-01-01", periods=3, freq="s"))
    >>> seconds_series
    0   2000-01-01 00:00:00
    1   2000-01-01 00:00:01
    2   2000-01-01 00:00:02
    dtype: datetime64[ns]
    >>> seconds_series.dt.second
    0    0
    1    1
    2    2
    dtype: int64

    >>> hours_series = pd.Series(pd.date_range("2000-01-01", periods=3, freq="h"))
    >>> hours_series
    0   2000-01-01 00:00:00
    1   2000-01-01 01:00:00
    2   2000-01-01 02:00:00
    dtype: datetime64[ns]
    >>> hours_series.dt.hour
    0    0
    1    1
    2    2
    dtype: int64

    >>> quarters_series = pd.Series(pd.date_range("2000-01-01", periods=3, freq="q"))
    >>> quarters_series
    0   2000-03-31
    1   2000-06-30
    2   2000-09-30
    dtype: datetime64[ns]
    >>> quarters_series.dt.quarter
    0    1
    1    2
    2    3
    dtype: int64

    Returns a Series indexed like the original Series.
    Raises TypeError if the Series does not contain datetimelike values.
    """
    def to_pydatetime(self) -> np.ndarray:
        """
        Return the data as an array of :class:`datetime.datetime` objects.

        Timezone information is retained if present.

        .. warning::

           Python's datetime uses microsecond resolution, which is lower than
           pandas (nanosecond). The values are truncated.

        Returns
        -------
        numpy.ndarray
            Object dtype array containing native Python datetime objects.

        See Also
        --------
        datetime.datetime : Standard library value for a datetime.

        Examples
        --------
        >>> s = pd.Series(pd.date_range('20180310', periods=2))
        >>> s
        0   2018-03-10
        1   2018-03-11
        dtype: datetime64[ns]

        >>> s.dt.to_pydatetime()
        array([datetime.datetime(2018, 3, 10, 0, 0),
               datetime.datetime(2018, 3, 11, 0, 0)], dtype=object)

        pandas' nanosecond precision is truncated to microseconds.

        >>> s = pd.Series(pd.date_range('20180310', periods=2, freq='ns'))
        >>> s
        0   2018-03-10 00:00:00.000000000
        1   2018-03-10 00:00:00.000000001
        dtype: datetime64[ns]

        >>> s.dt.to_pydatetime()
        array([datetime.datetime(2018, 3, 10, 0, 0),
               datetime.datetime(2018, 3, 10, 0, 0)], dtype=object)
        """
        ...
    
    @property
    def freq(self): # -> str | None:
        ...
    
    def isocalendar(self) -> DataFrame:
        """
        Calculate year, week, and day according to the ISO 8601 standard.

        .. versionadded:: 1.1.0

        Returns
        -------
        DataFrame
            With columns year, week and day.

        See Also
        --------
        Timestamp.isocalendar : Function return a 3-tuple containing ISO year,
            week number, and weekday for the given Timestamp object.
        datetime.date.isocalendar : Return a named tuple object with
            three components: year, week and weekday.

        Examples
        --------
        >>> ser = pd.to_datetime(pd.Series(["2010-01-01", pd.NaT]))
        >>> ser.dt.isocalendar()
           year  week  day
        0  2009    53     5
        1  <NA>  <NA>  <NA>
        >>> ser.dt.isocalendar().week
        0      53
        1    <NA>
        Name: week, dtype: UInt32
        """
        ...
    
    @property
    def weekofyear(self): # -> Any:
        """
        The week ordinal of the year according to the ISO 8601 standard.

        .. deprecated:: 1.1.0

        Series.dt.weekofyear and Series.dt.week have been deprecated.  Please
        call :func:`Series.dt.isocalendar` and access the ``week`` column
        instead.
        """
        ...
    
    week = ...


@delegate_names(delegate=TimedeltaArray, accessors=TimedeltaArray._datetimelike_ops, typ="property")
@delegate_names(delegate=TimedeltaArray, accessors=TimedeltaArray._datetimelike_methods, typ="method")
class TimedeltaProperties(Properties):
    """
    Accessor object for datetimelike properties of the Series values.

    Returns a Series indexed like the original Series.
    Raises TypeError if the Series does not contain datetimelike values.

    Examples
    --------
    >>> seconds_series = pd.Series(
    ...     pd.timedelta_range(start="1 second", periods=3, freq="S")
    ... )
    >>> seconds_series
    0   0 days 00:00:01
    1   0 days 00:00:02
    2   0 days 00:00:03
    dtype: timedelta64[ns]
    >>> seconds_series.dt.seconds
    0    1
    1    2
    2    3
    dtype: int64
    """
    def to_pytimedelta(self) -> np.ndarray:
        """
        Return an array of native :class:`datetime.timedelta` objects.

        Python's standard `datetime` library uses a different representation
        timedelta's. This method converts a Series of pandas Timedeltas
        to `datetime.timedelta` format with the same length as the original
        Series.

        Returns
        -------
        numpy.ndarray
            Array of 1D containing data with `datetime.timedelta` type.

        See Also
        --------
        datetime.timedelta : A duration expressing the difference
            between two date, time, or datetime.

        Examples
        --------
        >>> s = pd.Series(pd.to_timedelta(np.arange(5), unit="d"))
        >>> s
        0   0 days
        1   1 days
        2   2 days
        3   3 days
        4   4 days
        dtype: timedelta64[ns]

        >>> s.dt.to_pytimedelta()
        array([datetime.timedelta(0), datetime.timedelta(days=1),
        datetime.timedelta(days=2), datetime.timedelta(days=3),
        datetime.timedelta(days=4)], dtype=object)
        """
        ...
    
    @property
    def components(self):
        """
        Return a Dataframe of the components of the Timedeltas.

        Returns
        -------
        DataFrame

        Examples
        --------
        >>> s = pd.Series(pd.to_timedelta(np.arange(5), unit='s'))
        >>> s
        0   0 days 00:00:00
        1   0 days 00:00:01
        2   0 days 00:00:02
        3   0 days 00:00:03
        4   0 days 00:00:04
        dtype: timedelta64[ns]
        >>> s.dt.components
           days  hours  minutes  seconds  milliseconds  microseconds  nanoseconds
        0     0      0        0        0             0             0            0
        1     0      0        0        1             0             0            0
        2     0      0        0        2             0             0            0
        3     0      0        0        3             0             0            0
        4     0      0        0        4             0             0            0
        """
        ...
    
    @property
    def freq(self): # -> str | None:
        ...
    


@delegate_names(delegate=PeriodArray, accessors=PeriodArray._datetimelike_ops, typ="property")
@delegate_names(delegate=PeriodArray, accessors=PeriodArray._datetimelike_methods, typ="method")
class PeriodProperties(Properties):
    """
    Accessor object for datetimelike properties of the Series values.

    Returns a Series indexed like the original Series.
    Raises TypeError if the Series does not contain datetimelike values.

    Examples
    --------
    >>> seconds_series = pd.Series(
    ...     pd.period_range(
    ...         start="2000-01-01 00:00:00", end="2000-01-01 00:00:03", freq="s"
    ...     )
    ... )
    >>> seconds_series
    0    2000-01-01 00:00:00
    1    2000-01-01 00:00:01
    2    2000-01-01 00:00:02
    3    2000-01-01 00:00:03
    dtype: period[S]
    >>> seconds_series.dt.second
    0    0
    1    1
    2    2
    3    3
    dtype: int64

    >>> hours_series = pd.Series(
    ...     pd.period_range(start="2000-01-01 00:00", end="2000-01-01 03:00", freq="h")
    ... )
    >>> hours_series
    0    2000-01-01 00:00
    1    2000-01-01 01:00
    2    2000-01-01 02:00
    3    2000-01-01 03:00
    dtype: period[H]
    >>> hours_series.dt.hour
    0    0
    1    1
    2    2
    3    3
    dtype: int64

    >>> quarters_series = pd.Series(
    ...     pd.period_range(start="2000-01-01", end="2000-12-31", freq="Q-DEC")
    ... )
    >>> quarters_series
    0    2000Q1
    1    2000Q2
    2    2000Q3
    3    2000Q4
    dtype: period[Q-DEC]
    >>> quarters_series.dt.quarter
    0    1
    1    2
    2    3
    3    4
    dtype: int64
    """
    ...


class CombinedDatetimelikeProperties(DatetimeProperties, TimedeltaProperties, PeriodProperties):
    def __new__(cls, data: Series): # -> DatetimeProperties | TimedeltaProperties | PeriodProperties:
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from numpy import typing as npt

class NAType:
    def __new__(cls, *args, **kwargs):
        ...
    


NA: NAType
def is_matching_na(left: object, right: object, nan_matches_none: bool = ...) -> bool:
    ...

def isposinf_scalar(val: object) -> bool:
    ...

def isneginf_scalar(val: object) -> bool:
    ...

def checknull(val: object, inf_as_na: bool = ...) -> bool:
    ...

def isnaobj(arr: np.ndarray, inf_as_na: bool = ...) -> npt.NDArray[np.bool_]:
    ...

def isnaobj2d(arr: np.ndarray, inf_as_na: bool = ...) -> npt.NDArray[np.bool_]:
    ...

def is_numeric_na(values: np.ndarray) -> npt.NDArray[np.bool_]:
    ...

def is_float_nan(values: np.ndarray) -> npt.NDArray[np.bool_]:
    ...




"""
This type stub file was generated by pyright.
"""

def show_versions(as_json: str | bool = ...) -> None:
    """
    Provide useful information, important for bug reports.

    It comprises info about hosting operation system, pandas version,
    and versions of other installed relative packages.

    Parameters
    ----------
    as_json : str or bool, default False
        * If False, outputs info in a human readable form to the console.
        * If str, it will be considered as a path to a file.
          Info will be written to that file in JSON format.
        * If True, outputs info in JSON format to the console.
    """
    ...

def main() -> int:
    ...

if __name__ == "__main__":
    ...



"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any
from pandas._typing import ArrayLike, Shape

"""
Functions for arithmetic and comparison operations on NumPy arrays and
ExtensionArrays.
"""
def comp_method_OBJECT_ARRAY(op, x, y): # -> ndarray[Any, dtype[bool_]]:
    ...

def arithmetic_op(left: ArrayLike, right: Any, op): # -> ndarray[Unknown, Unknown] | tuple[ndarray[Unknown, Unknown], Unknown]:
    """
    Evaluate an arithmetic operation `+`, `-`, `*`, `/`, `//`, `%`, `**`, ...

    Note: the caller is responsible for ensuring that numpy warnings are
    suppressed (with np.errstate(all="ignore")) if needed.

    Parameters
    ----------
    left : np.ndarray or ExtensionArray
    right : object
        Cannot be a DataFrame or Index.  Series is *not* excluded.
    op : {operator.add, operator.sub, ...}
        Or one of the reversed variants from roperator.

    Returns
    -------
    ndarray or ExtensionArray
        Or a 2-tuple of these in the case of divmod or rdivmod.
    """
    ...

def comparison_op(left: ArrayLike, right: Any, op) -> ArrayLike:
    """
    Evaluate a comparison operation `=`, `!=`, `>=`, `>`, `<=`, or `<`.

    Note: the caller is responsible for ensuring that numpy warnings are
    suppressed (with np.errstate(all="ignore")) if needed.

    Parameters
    ----------
    left : np.ndarray or ExtensionArray
    right : object
        Cannot be a DataFrame, Series, or Index.
    op : {operator.eq, operator.ne, operator.gt, operator.ge, operator.lt, operator.le}

    Returns
    -------
    ndarray or ExtensionArray
    """
    ...

def na_logical_op(x: np.ndarray, y, op): # -> ndarray[Any, Unknown]:
    ...

def logical_op(left: ArrayLike, right: Any, op) -> ArrayLike:
    """
    Evaluate a logical operation `|`, `&`, or `^`.

    Parameters
    ----------
    left : np.ndarray or ExtensionArray
    right : object
        Cannot be a DataFrame, Series, or Index.
    op : {operator.and_, operator.or_, operator.xor}
        Or one of the reversed variants from roperator.

    Returns
    -------
    ndarray or ExtensionArray
    """
    ...

def get_array_op(op): # -> partial[Unknown] | partial[ArrayLike] | partial[Unknown | ndarray[Unknown, Unknown] | tuple[ndarray[Unknown, Unknown], Unknown]]:
    """
    Return a binary array operation corresponding to the given operator op.

    Parameters
    ----------
    op : function
        Binary operator from operator or roperator module.

    Returns
    -------
    functools.partial
    """
    ...

def maybe_prepare_scalar_for_op(obj, shape: Shape): # -> Timedelta | NaTType | Timestamp:
    """
    Cast non-pandas objects to pandas types to unify behavior of arithmetic
    and comparison operations.

    Parameters
    ----------
    obj: object
    shape : tuple[int]

    Returns
    -------
    out : object

    Notes
    -----
    Be careful to call this *after* determining the `name` attribute to be
    attached to the result of the arithmetic operation.
    """
    ...

_BOOL_OP_NOT_ALLOWED = ...



"""
This type stub file was generated by pyright.
"""

from typing import Hashable
from pandas._typing import Dtype
from pandas.core.frame import DataFrame

def get_dummies(data, prefix=..., prefix_sep=..., dummy_na: bool = ..., columns=..., sparse: bool = ..., drop_first: bool = ..., dtype: Dtype | None = ...) -> DataFrame:
    """
    Convert categorical variable into dummy/indicator variables.

    Parameters
    ----------
    data : array-like, Series, or DataFrame
        Data of which to get dummy indicators.
    prefix : str, list of str, or dict of str, default None
        String to append DataFrame column names.
        Pass a list with length equal to the number of columns
        when calling get_dummies on a DataFrame. Alternatively, `prefix`
        can be a dictionary mapping column names to prefixes.
    prefix_sep : str, default '_'
        If appending prefix, separator/delimiter to use. Or pass a
        list or dictionary as with `prefix`.
    dummy_na : bool, default False
        Add a column to indicate NaNs, if False NaNs are ignored.
    columns : list-like, default None
        Column names in the DataFrame to be encoded.
        If `columns` is None then all the columns with
        `object`, `string`, or `category` dtype will be converted.
    sparse : bool, default False
        Whether the dummy-encoded columns should be backed by
        a :class:`SparseArray` (True) or a regular NumPy array (False).
    drop_first : bool, default False
        Whether to get k-1 dummies out of k categorical levels by removing the
        first level.
    dtype : dtype, default np.uint8
        Data type for new columns. Only a single dtype is allowed.

    Returns
    -------
    DataFrame
        Dummy-coded data.

    See Also
    --------
    Series.str.get_dummies : Convert Series to dummy codes.
    :func:`~pandas.from_dummies` : Convert dummy codes to categorical ``DataFrame``.

    Notes
    -----
    Reference :ref:`the user guide <reshaping.dummies>` for more examples.

    Examples
    --------
    >>> s = pd.Series(list('abca'))

    >>> pd.get_dummies(s)
       a  b  c
    0  1  0  0
    1  0  1  0
    2  0  0  1
    3  1  0  0

    >>> s1 = ['a', 'b', np.nan]

    >>> pd.get_dummies(s1)
       a  b
    0  1  0
    1  0  1
    2  0  0

    >>> pd.get_dummies(s1, dummy_na=True)
       a  b  NaN
    0  1  0    0
    1  0  1    0
    2  0  0    1

    >>> df = pd.DataFrame({'A': ['a', 'b', 'a'], 'B': ['b', 'a', 'c'],
    ...                    'C': [1, 2, 3]})

    >>> pd.get_dummies(df, prefix=['col1', 'col2'])
       C  col1_a  col1_b  col2_a  col2_b  col2_c
    0  1       1       0       0       1       0
    1  2       0       1       1       0       0
    2  3       1       0       0       0       1

    >>> pd.get_dummies(pd.Series(list('abcaa')))
       a  b  c
    0  1  0  0
    1  0  1  0
    2  0  0  1
    3  1  0  0
    4  1  0  0

    >>> pd.get_dummies(pd.Series(list('abcaa')), drop_first=True)
       b  c
    0  0  0
    1  1  0
    2  0  1
    3  0  0
    4  0  0

    >>> pd.get_dummies(pd.Series(list('abc')), dtype=float)
         a    b    c
    0  1.0  0.0  0.0
    1  0.0  1.0  0.0
    2  0.0  0.0  1.0
    """
    ...

def from_dummies(data: DataFrame, sep: None | str = ..., default_category: None | Hashable | dict[str, Hashable] = ...) -> DataFrame:
    """
    Create a categorical ``DataFrame`` from a ``DataFrame`` of dummy variables.

    Inverts the operation performed by :func:`~pandas.get_dummies`.

    .. versionadded:: 1.5.0

    Parameters
    ----------
    data : DataFrame
        Data which contains dummy-coded variables in form of integer columns of
        1's and 0's.
    sep : str, default None
        Separator used in the column names of the dummy categories they are
        character indicating the separation of the categorical names from the prefixes.
        For example, if your column names are 'prefix_A' and 'prefix_B',
        you can strip the underscore by specifying sep='_'.
    default_category : None, Hashable or dict of Hashables, default None
        The default category is the implied category when a value has none of the
        listed categories specified with a one, i.e. if all dummies in a row are
        zero. Can be a single value for all variables or a dict directly mapping
        the default categories to a prefix of a variable.

    Returns
    -------
    DataFrame
        Categorical data decoded from the dummy input-data.

    Raises
    ------
    ValueError
        * When the input ``DataFrame`` ``data`` contains NA values.
        * When the input ``DataFrame`` ``data`` contains column names with separators
          that do not match the separator specified with ``sep``.
        * When a ``dict`` passed to ``default_category`` does not include an implied
          category for each prefix.
        * When a value in ``data`` has more than one category assigned to it.
        * When ``default_category=None`` and a value in ``data`` has no category
          assigned to it.
    TypeError
        * When the input ``data`` is not of type ``DataFrame``.
        * When the input ``DataFrame`` ``data`` contains non-dummy data.
        * When the passed ``sep`` is of a wrong data type.
        * When the passed ``default_category`` is of a wrong data type.

    See Also
    --------
    :func:`~pandas.get_dummies` : Convert ``Series`` or ``DataFrame`` to dummy codes.
    :class:`~pandas.Categorical` : Represent a categorical variable in classic.

    Notes
    -----
    The columns of the passed dummy data should only include 1's and 0's,
    or boolean values.

    Examples
    --------
    >>> df = pd.DataFrame({"a": [1, 0, 0, 1], "b": [0, 1, 0, 0],
    ...                    "c": [0, 0, 1, 0]})

    >>> df
       a  b  c
    0  1  0  0
    1  0  1  0
    2  0  0  1
    3  1  0  0

    >>> pd.from_dummies(df)
    0     a
    1     b
    2     c
    3     a

    >>> df = pd.DataFrame({"col1_a": [1, 0, 1], "col1_b": [0, 1, 0],
    ...                    "col2_a": [0, 1, 0], "col2_b": [1, 0, 0],
    ...                    "col2_c": [0, 0, 1]})

    >>> df
          col1_a  col1_b  col2_a  col2_b  col2_c
    0       1       0       0       1       0
    1       0       1       1       0       0
    2       1       0       0       0       1

    >>> pd.from_dummies(df, sep="_")
        col1    col2
    0    a       b
    1    b       a
    2    a       c

    >>> df = pd.DataFrame({"col1_a": [1, 0, 0], "col1_b": [0, 1, 0],
    ...                    "col2_a": [0, 1, 0], "col2_b": [1, 0, 0],
    ...                    "col2_c": [0, 0, 0]})

    >>> df
          col1_a  col1_b  col2_a  col2_b  col2_c
    0       1       0       0       1       0
    1       0       1       1       0       0
    2       0       0       0       0       0

    >>> pd.from_dummies(df, sep="_", default_category={"col1": "d", "col2": "e"})
        col1    col2
    0    a       b
    1    b       a
    2    d       e
    """
    ...




"""
This type stub file was generated by pyright.
"""

"""
Functions to generate methods and pin them to the appropriate classes.
"""
def add_flex_arithmetic_methods(cls) -> None:
    """
    Adds the full suite of flex arithmetic methods (``pow``, ``mul``, ``add``)
    to the class.

    Parameters
    ----------
    cls : class
        flex methods will be defined and pinned to this class
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numba
import numpy as np

@numba.jit(numba.boolean(numba.int64[:]), nopython=True, nogil=True, parallel=False)
def is_monotonic_increasing(bounds: np.ndarray) -> bool:
    """Check if int64 values are monotonically increasing."""
    ...




"""
This type stub file was generated by pyright.
"""

"""
Public testing utility functions.
"""
__all__ = ["assert_extension_array_equal", "assert_frame_equal", "assert_series_equal", "assert_index_equal"]



"""
This type stub file was generated by pyright.
"""

from pandas._config import config, dates
from pandas._config.config import describe_option, get_option, option_context, options, reset_option, set_option
from pandas._config.display import detect_console_encoding

"""
pandas._config is considered explicitly upstream of everything else in pandas,
should have no intra-pandas dependencies.

importing `dates` and `display` ensures that keys needed by _libs
are initialized.
"""
__all__ = ["config", "detect_console_encoding", "get_option", "set_option", "reset_option", "describe_option", "option_context", "options"]
